# Python version 3.5 and up.

from setuptools import setup, find_packages
from codecs import open
from os import path


here = path.abspath(path.dirname(__file__))
install_reqs = None
with open('requirements.txt') as f:
    install_reqs = f.read().splitlines()

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    version='2.7.7',
    name='p2pd',
    description='Asynchronous P2P networking library and service',
    keywords=('NAT traversal, TCP hole punching, simultaneous open, UPnP, STUN, TURN, SIP, DHCP, add IP to interface, NATPMP, P2P, Peer-to-peer networking library, python'),
    long_description_content_type="text/markdown",
    long_description=long_description,
    url='http://github.com/robertsdotpm/p2pd',
    author='Matthew Roberts',
    author_email='matthew@roberts.pm',
    license='public domain',
    package_dir={"": "."},
    packages=find_packages(exclude=('tests', 'docs')),
    package_data={'p2pd': ['scripts/kvs_schema.sqlite3']},
    include_package_data=True,
    install_requires=install_reqs,
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3'
    ],
)



"""
- Firewalls for IRC servers silently drop packets if you
make successive connections too closely.
- When registration of channel names is done a user must first join
the channel. The problem is: the moment someone joins a channel the name
of the channel becomes public. That means that an attacker can watch the
channel list for channels and register channel names on others servers
in order to hijack the control of names before registering parties.

The solution I came up to this is multi-faceted:
    (1) Each name is now server-specific. By hashing the server-name as
    part of the name it will be harder to guess what name is being
    registered to register channels at other servers.
    (2) The names contain a small proof-of-work that acts as a delay
    function. The registering party 'saves up' these proofs for all
    servers in bulk and then registers everything in parallel. Naturally,
    this becomes part of the hashed name.

But what about rainbow tables?
    (1) Names may be registered on arbitrary TLDs. Each TLD acts as
    a small salt. Thus, attackers have to guess what TLDs to
    pre-compute which may be completely off. If TLD suggestions are
    randomly generated from dictionary nouns, then rainbow tables
    would be near impossible to create.
    (2) Names may contain an optional password portion. This has the
    downside of requiring the password to access the name but offers
    more protection against squatters.

These measures make race conditions and rainbow tables either
impossible or impractical depending on usage.

---
    high priority:
    - finish writing article
    - merge to main and push to pypi
    - post article
    
----------------------------
    future features out of scope:

    - need a mechanism to fetch active servers from a repo if servers
    have registration disabled

    - if account gets deleted for inactivity?
        - refresh to prevent that
        - channel successor
        - if an operator is interfering then could disable that server frm the list automatically (and the account should stil be active)

--------------------------------------
"""

import asyncio
import os
import re
import random
import time
import struct
import argon2pure
import binascii
from concurrent.futures import ProcessPoolExecutor
from ecdsa import SigningKey, VerifyingKey, SECP256k1
from .utils import *
from .address import *
from .interface import *
from .base_stream import *
from .base_n import encodebytes, decodebytes
from .base_n import B36_CHARSET, B64_CHARSET, B92_CHARSET
from .install import *
from .sqlite_kvs import *

IRC_VERSION = "Friendly P2PD user - see p2pd.net/irc"

IRC_CHAN_REFRESH = "Refreshing chan as we're still using this! See p2pd.net/irc for more information."

IRC_NICK_REFRESH = "Refreshing nick as we're still using this!"

IRC_SERVERS = [
    # Works.
    {
        'domain': 'irc.phat-net.de',
        'afs': [IP4, IP6],

        # 6 nov 2000
        'creation': 975848400,

        'nick_serv': ["password", "email"],

        "ip": {
            IP4: "116.203.29.246",
            IP6: "2a01:4f8:c2c:628::1"
        },

        'chan_len': 50,
        'chan_no': 60,
        'topic_len': 300,
        'chan_topics': 'a-zA-Z0-9all specials unicode (smilies tested)',
        'nick_expiry': 21,
        'chan_expiry': 14,
        'test_chan': '#test',
        "successor": "p2pd_matthew",
        'unregistered_join': True,
    },
    # Works.
    {
        'domain': 'irc.tweakers.net',
        'afs': [IP4, IP6],

        # 30 apr 2002
        'creation': 1020088800,

        'chan_serv': ["password", "description"],
        'nick_serv': ["password", "email"],

        "ip": {
            IP4: "213.239.154.35",
            IP6: "2001:9a8:0:e:1337::6667" # Top kek.
        },

        'chan_no': 25,
        'chan_len': 32,
        'topic_len': 307,
        'chan_topics': 'a-zA-Z0-9all specials unicode (smilies)',
        'nick_expiry': 210,
        'chan_expiry': 21,
        'test_chan': '#test',
        "successor": "p2pd_matthew",
        'unregistered_join': True,
    },
    {
        'domain': 'irc.swiftirc.net',
        'afs': [IP4, IP6],

        # 10 march 2007
        'creation': 1173445200,

        'nick_serv': ["password", "email"],
        #'set_topic': "set",

        "ip": {
            IP4: "159.65.55.232",
            IP6: "2604:a880:4:1d0::75:0" # Top kek.
        },

        'chan_no': 50,
        'chan_len': 32,
        'topic_len': 360,

        # Certain characters like ,. still not allowed.
        'chan_name': 'azAZ!@#$unicode-_',
        'chan_topics': 'a-zA-Z0-9all specials unicode (smilies)',
        'nick_expiry': 90,
        'chan_expiry': 14,
        'test_chan': '#mSL.test',
        "successor": "p2pd_matthew",
        'unregistered_join': True,
    },
    {
        'domain': 'irc.euirc.net',
        'afs': [IP4, IP6],

        # 19 sep 2000
        "creation": 969282000,
        'chan_serv': ["password", "description"],
        'nick_serv': ["password", "email"],
        'set_topic': "set",

        "ip": {
            IP4: "83.137.40.10",
            IP6: "2001:41d0:701:1000::9b"
        },

        'chan_no': 30,
        'chan_len': 32,
        'topic_len': 307,
        'chan_topics': 'a-zA-Z0-9all specials unicode (smilies)',
        'nick_expiry': 120,
        'chan_expiry': 60,
        'test_chan': '#test',
        'unregistered_join': True,
        "successor": "p2pd_matthew"
    },
    {
        'domain': 'irc.darkmyst.org',
        'afs': [IP4, IP6],

        # 26 nov 2002
        'creation': 1038229200,
        'nick_serv': ["password", "email"],

        "ip": {
            IP4: "23.239.26.75",
            IP6: "2604:a880:cad:d0::1d:e001"
        },

        'chan_no': 30,
        'chan_len': 50,
        'topic_len': 390,
        'chan_topics': 'a-zA-Z0-9all specials unicode (smilies)',
        'nick_expiry': 30,
        'chan_expiry': 60,
        'test_chan': '#test',
        'unregistered_join': True,
        "successor": "p2pd_matthew"
    },
]

###################################################################
# These are really IRC-server specific response strings.
IRC_POS_LOGIN_MSGS = [
    "now logged in as",
    "now identified for",
    "with the password",
    "you are now recognized",
    "Nickname [^\s]* registered"
]

IRC_POS_CHAN_MSGS = [
    "registered under",
    "is now registered",
    "successfully registered",
]

IRC_NEG_CHAN_MSGS = [
    "not complete registration",
    "following link",
    "link expires",
    "address is not confirmed"
]

IRC_NICK_POS = [
    "remember this for later",
    "nickname is registered",
    "ickname \S* is already",
    #"is reserved by a different account"
]

IRC_REGISTER_SUCCESS = 0
IRC_START_FAILURE = 1
IRC_REGISTER_FAILURE = 2
IRC_REGISTER_PARTIAL = 3

IRC_CONF = dict_child({
    "use_ssl": 1,
    "ssl_handshake": 8,
    "con_timeout": 4,
}, NET_CONF)

IRC_FRESHNESS = 172800

IRC_PREFIX = b"20"
IRC_SALT = b"IRCDNS - improve entropy for weak seeds."

####################################################################################

# SHA3 digest in ascii truncated to a str limit.
# Used for deterministic passwords with good complexity.
def f_irc_pass(x):
    return to_s(
        encodebytes(
            hashlib.sha3_256(
                to_b(x)
            ).digest(),
            charset=B64_CHARSET
        )
    )[:30]

# Used for encoding channel names.
def f_sha3_b36(msg):
    h_b = hashlib.sha3_256(to_b(msg)).digest()
    return to_s(
        encodebytes(
            h_b,
            charset=B36_CHARSET
        )
    )

# Used to make sure that ECDSA private keys are valid when using hashes for them.
def f_sha3_to_ecdsa_priv(msg):
    # Numbers must be <= to this hex number.
    max_hex = "FFFFFFFF00000000FFFFFFFFFFFFFFFF"
    max_hex += "BCE6FAADA7179E84F3B9CAC2FC632551"
    max_int = int(max_hex, 16)

    # Function hashes the input until it meets the requirements above.
    # Continue until hash hex is <= max_int.
    while 1:
        h_b = hashlib.sha3_256(to_b(msg)).digest()
        h_hex = binascii.hexlify(h_b)
        h_int = int(h_hex, 16)
        if h_int <= max_int:
            return h_b
        else:
            msg = h_b

# Function that adds verifiable delays to channel names to
# prevent race conditions when registering channel names.
def f_chan_pow(msg):
    # Time locks help prevent registration front-running.
    # Argon2 is highly adjustable for memory, CPU, and GPUs.
    time_lock = argon2pure.argon2(
        # Msg portion.
        msg,

        # Unique salt for rainbow resistance.
        b'Ana main, chicken tikka masala, Blush - Wolf Alice',
        
        # Iterations.
        time_cost=2, # 700
        
        # KB needed.
        memory_cost=1024 * 2,
        
        # Threads needed.
        parallelism=1
    )

    # Channel names begin with a #.
    h = "#" + to_s(
        # The result is encoded using A-Z0-9 for chan names.
        encodebytes(
            # Multiple hash functions make collisions harder to find.
            # As a value will need to work for both functions.
            hash160(
                hashlib.sha3_256(
                    msg + time_lock
                ).digest()
            ),
            charset=B36_CHARSET
        )
    )[:31].lower()

    # Cache.
    return h, time_lock

# Not meant to provide bullet-proof security.
# A simple function that implements one-time pads with key-stretching.
def f_otp(msg, otp):
    # Make otp long enough.
    while len(otp) < len(msg):
        otp += b_sha3_256(otp)

    # Xor MSG with OTP.
    buf = b""
    for i in range(0, len(msg)):
        buf += bytes([
            msg[i] ^ otp[i]
        ])

    return buf

# IRC channel topics can be used to store a value.
# My algorithm uses threshold consensus to choose a topic value.
# Signed, encrypted, and secured against race conditions.
async def f_pack_topic(value, name, tld, pw, ses, clsChan, executor=None):
    # Bytes used to make ECDSA priv key.
    priv = f_sha3_to_ecdsa_priv(
        to_b(f"{name}{tld}{pw}") + IRC_SALT + ses.seed
    )

    # ECDSA secret key.
    sk = SigningKey.from_string(
        string=priv,
        hashfunc=hashlib.sha3_256,
        curve=SECP256k1
    )

    # Generate channel name.
    chan_name = await ses.get_irc_chan_name(
        name,
        tld,
        pw,
        executor
    )

    # Load channel manager.
    if chan_name not in ses.chans:
        chan = clsChan(chan_name, ses)
        ses.chans[chan_name] = chan
    else:
        chan = ses.chans[chan_name]

    # Time lock value used for OTP.
    time_lock = ses.chan_time_locks[chan_name]

    # Serialized data portion with timestamp prepend.
    val_buf = struct.pack("Q", int(time.time()))
    val_buf += to_b(value)

    # Weakly encrypted topic value.
    # The payload portion has a unique one-time pad from the time-lock.
    val_buf = f_otp(
        val_buf,
        b_sha3_256(b"msg " + time_lock)
    )

    # Signed val_buf with vk.
    # Signal portion has a unique one-time pad from the time-lock.
    sig_buf = sk.sign(val_buf)
    sig_buf = f_otp(
        sig_buf,
        b_sha3_256(b"sig " + time_lock)
    )

    # Topic message to store (topic-safe encoding.)
    return "%s %s %s" % (
        "p2pd.net/irc",
        to_s(encodebytes(sig_buf, charset=B92_CHARSET)),
        to_s(encodebytes(val_buf, charset=B92_CHARSET)),
    ), chan

# This function validates the format of a topic.
# It will extract ECDSA public keys from the signature.
# Validate the signature across the message.
# Handle base decode and decryption.
def f_unpack_topic(chan_name, topic, session):
    if topic is None:
        return None
    
    parts = topic.split()
    if len(parts) != 3:
        return None
    
    # Time lock value used for OTP.
    time_lock = session.chan_time_locks[chan_name]

    # Signature part.
    sig_b = decodebytes(parts[1], charset=B92_CHARSET)
    sig_b = f_otp(
        sig_b,
        b_sha3_256(b"sig " + time_lock)
    )

    # Message part.
    msg_b = decodebytes(parts[2], charset=B92_CHARSET)

    # List of possible public keys recovered from sig.
    vk_list = VerifyingKey.from_public_key_recovery(
        signature=sig_b,
        data=msg_b,
        curve=SECP256k1,
        hashfunc=hashlib.sha3_256
    )

    # Public key extraction returns multiple possibilities.
    # The algorithm converges on the right key through basic consensus.
    vk_ids = []
    for vk in vk_list:
        vk_ids.append(
            vk.to_string("compressed")
        )

    # Anything that validates is a public key worth keeping.
    for vk in vk_list:
        try:
            vk.verify(
                sig_b,
                msg_b
            )

            msg_b = f_otp(msg_b, b_sha3_256(b"msg " + time_lock))
            timestamp = struct.unpack("Q", msg_b[:8])[0]
            return {
                "ids": vk_ids,
                "vk": vk,
                "msg": to_s(msg_b[8:]),
                "otp": f_otp(msg_b, b_sha3_256(b"msg " + time_lock)),
                "sig": sig_b,
                "time": timestamp,
            }
        except:
            log_exception()

# Matches a basic channel name in IRC.
def irc_is_valid_chan_name(chan_name):
    p = "^[#&+!][a-zA-Z0-9_-]+$"
    return re.match(p, chan_name) != None

# Example user!ident@hostname.
# Used to validate senders of messages from IRC protocol messages.
def irc_extract_sender(sender):
    p = "([^!@]+)(?:!([^@]+))?(?:@([\s\S]+))?"
    parts = re.findall(p, sender)
    if len(parts):
        parts = parts[0]
        return {
            "nick": parts[0],
            "user": parts[1],
            "host": parts[2]
        }
    
    raise Exception("Sender portion invalid.")

"""
TCP is stream-oriented and portions of IRC protocol messages
may be received. This code handles finding valid IRC messages and
truncating the recv buffer once extracted.
"""
def irc_extract_msgs(buf):
    #     optional                                       optional
    #     :prefix            CMD           param...      :suffix 
    # Changed so that multiple spaces can sep portions.
    p = "(?:[:]([^ :]+?) +)?([A-Z0-9]+) +(?:([^\r\:]+?) *)?(?:[:]([^\r]+))?\r\n"
    p = re.compile(p)

    # Loop over all protocol messages in buf.
    # Truncate buf as messages are extracted.
    msgs = []
    offset = None
    while 1:
        # No matches found yet.
        # Buffer contains a partial response or its empty.
        match = p.search(buf)
        if match is None:
            break

        # Extract matched message.
        msg = IRCMsg(*list(match.groups()))
        msgs.append(msg)

        # Get the offset of the match end.
        # Truncate buf as messages are extracted.
        offset = match.end()
        if offset < len(buf):
            buf = buf[offset:]
        else:
            buf = ""

    return msgs, buf

"""
IRC is a simple text-bases protocol. This class handles building valid
IRC messages for sending replies back to the server.
Many parts of the IRC protocol message are optional and this class handles that.
"""
class IRCMsg():
    def __init__(self, prefix="", cmd="", param="", suffix=""):
        irc_safe = lambda m: m.replace(":", "")
        self.prefix = irc_safe(prefix or "")
        self.cmd = irc_safe(cmd or "")
        self.param = irc_safe(param or "")
        self.suffix = suffix or ""

    # Build a valid IRC protocol message.
    def pack(self):
        out = ""
        if len(self.prefix):
            out += f":{self.prefix} "

        out += self.cmd
        if len(self.param):
            out += f" {self.param}"
        
        if len(self.suffix):
            out += f" :{self.suffix}"

        out += "\r\n"
        return to_b(out)
    
    # Output the protocol message as a string.
    def __str__(self):
        return to_s(self.pack())
    
    # Output the protocol message as bytes.
    def __bytes__(self):
        return self.pack()
    
    # Compare the serialized version of two IRCMsg classes.
    def __eq__(self, other):
        return str(self) == str(other)

"""
IRC channels need to know how to encode their own names
and keep their topics updated.
"""
class IRCChan:
    def __init__(self, chan_name, session):
        self.session = session
        self.chan_name = chan_name
        self.chan_pass = f_irc_pass(
            IRC_PREFIX +
            to_b(chan_name) + 
            to_b(session.irc_server) + 
            IRC_SALT + 
            session.seed
        )

        self.set_topic_done = asyncio.Future()
        self.pending_topic = ""
        
    async def get_ops(self):
        session = self.session
        session.chan_ident[self.chan_name] = asyncio.Future()

        # Join channel.
        await session.con.send(
            IRCMsg(
                cmd="JOIN",
                param=f"{self.chan_name}",
            ).pack()
        )

        await session.con.send(
            IRCMsg(
                cmd="PRIVMSG",
                param="ChanServ",
                suffix=f"OP {self.chan_name}"
            ).pack()
        )

        return await session.chan_ident[self.chan_name]
        
    async def set_topic(self, topic):
        self.pending_topic = topic
        session = self.session

        # Join channel.
        await session.con.send(
            IRCMsg(
                cmd="JOIN",
                param=f"{self.chan_name}",
            ).pack()
        )

        if 'set_topic' in session.server_info:
            if session.server_info['set_topic'] == 'set':
                await session.con.send(
                    IRCMsg(
                        cmd="PRIVMSG",
                        param="ChanServ",
                        suffix=f"SET {self.chan_name} TOPIC {topic}"
                    ).pack()
                )
        else:
            await session.con.send(
                IRCMsg(
                    cmd="PRIVMSG",
                    param="ChanServ",
                    suffix=f"TOPIC {self.chan_name} {topic}"
                ).pack()
            )

        await asyncio.wait_for(
            self.set_topic_done,
            10
        )

        self.set_topic_done = asyncio.Future()
        return self

"""
The sub-set of the IRC protocol this software implements is as bellow.
This function is purely data-bases. It takes IRCMsg objects and either
returns new IRCMsg objects as replies or None. The caller is responsible
for transmitting these replies. This greatly simplifies testing the
protocol as there is no network I/O done here.
"""
def irc_proto(self, msg):
    log(f"Got {msg.pack()}")

    # Process ping.
    if msg.cmd == "PING":
        return IRCMsg(
            cmd="PONG",
            param=msg.param,
            suffix=msg.suffix,
        )
    
    # End of motd.
    if msg.cmd in ["376", "411"]:
        self.get_motd.set_result(True)
        return

    # Nickname already reserved so login.
    if msg.cmd == "433":
        return IRCMsg(
            cmd="PRIVMSG",
            param="NickServ",
            suffix=f"IDENTIFY {self.user_pass}"
        )
    
    # Login if account already exists.
    for nick_success in IRC_NICK_POS:
        if msg.cmd != "NOTICE":
            continue

        if len(re.findall(nick_success, msg.suffix)):
            log("sending identify. 2")
            return IRCMsg(
                cmd="PRIVMSG",
                param="NickServ",
                suffix=f"IDENTIFY {self.user_pass}"
            )

    # Login success.
    if msg.cmd == "900":
        self.login_status.set_result(True)
        return

    # Login ident success or account register.
    for success_msg in IRC_POS_LOGIN_MSGS:
        if msg.cmd != "NOTICE":
            continue

        if len(re.findall(success_msg, msg.suffix)):
            log("login success")
            self.login_status.set_result(True)
            return

    # Respond to CTCP version requests.
    if msg.suffix == "\x01VERSION\x01":
        sender = irc_extract_sender(msg.prefix)
        return IRCMsg(
            cmd="PRIVMSG",
            param=sender["nick"],
            suffix=f"\x01VERSION {IRC_VERSION}\x01"
        )

    # Got a channels topic.
    if msg.cmd == "332":
        _, chan_part = msg.param.split()
        if chan_part in self.chan_topics:
            self.chan_topics[chan_part].set_result(msg.suffix)

    # Handle outstanding channel operations.
    for chan in d_keys(self.chans):
        # Process chan oper success.
        if msg.cmd == "MODE":
            op_success = f"{chan} +o {self.nick}".lower()
            if op_success in msg.param.lower():
                if chan in self.chan_ident:
                    self.chan_ident[chan].set_result(True)

        # Indicate topic set successfully.
        if msg.cmd == "TOPIC":
            if chan in msg.param:
                self.chans[chan].set_topic_done.set_result(
                    msg.suffix or True
                )

    # The rest must be in response to notices.
    if msg.cmd != "NOTICE":
        return

    # Support checking if a channel is registered.
    # Response for a channel INFO request.
    for chan_name in d_keys(self.chan_infos):
        # Channel is not registered.
        p = "annel \S*" + re.escape(chan_name) + "\S* ((isn)|(is not))"
        if len(re.findall(p, msg.suffix)):
            self.active_info_chan = chan_name
            self.chan_infos[chan_name].set_result(False)

        # Channel is registered.
        p = "mation ((for)|(on)) [^#]*" + re.escape(chan_name)
        if len(re.findall(p, msg.suffix)):
            self.active_info_chan = chan_name
        p = "annel \S*" + re.escape(chan_name) + "\S* is reg"
        if len(re.findall(p, msg.suffix)):
            self.active_info_chan = chan_name

    # Set channel owner.
    p = "[fF]ounder *[:] *([^:]+)"
    owner = re.findall(p, msg.suffix)
    if len(owner):
        if self.active_info_chan in self.chan_infos:
            if not self.chan_infos[self.active_info_chan].done():
                self.chan_infos[self.active_info_chan].set_result(
                    owner[0]
                )

"""
An 'IRCSession' is designed to record a connection to a unique IRC server.
These sessions have names uniquely identified to them through channel
registrations. The session needs to be able to implement basic features
of IRC like channel registration, user registration, login, and so on.
"""
class IRCSession():
    def __init__(self, server_info, seed, db=None, offset=None):
        assert(len(seed) >= 24)
        self.started = asyncio.Future()
        self.get_motd = asyncio.Future()
        self.login_status = asyncio.Future()
        self.chans_loaded = asyncio.Future()
        self.con = None
        self.recv_buf = ""
        self.chan_owners = {}
        self.chan_topics = {}
        self.chan_ident = {}
        self.chan_set_topic = {}
        self.chan_get_topic = {}
        self.chan_registered = {}
        self.chan_infos = {}; self.active_info_chan = None
        self.chan_name_hashes = {}
        self.chan_time_locks = {}
        self.server_info = server_info
        self.seed = seed
        self.db = db
        self.offset = offset

        # All IRC channels registered to this username.
        self.chans = {}

        # Derive details for IRC server.
        """
        Session details are determined by a combination of the unique
        IRC server's domain with the cryptographically secure seed value.
        This allows for password management to be far easier.
        """
        self.irc_server = self.server_info["domain"]
        self.username = "u" + f_sha3_b36(IRC_PREFIX + b":user:" + to_b(self.irc_server) + IRC_SALT + seed)[:7]
        self.user_pass = f_irc_pass(IRC_PREFIX + b":pass:" + to_b(self.irc_server) + IRC_SALT + seed)
        self.nick = "n" + f_sha3_b36(IRC_PREFIX + b":nick:" + to_b(self.irc_server) + IRC_SALT + seed)[:7]
        self.email = "e" + f_sha3_b36(IRC_PREFIX + b":email:" + to_b(self.irc_server) + IRC_SALT + seed)[:15]
        self.email += "@p2pd.net"

        # Sanity checks.
        assert(len(self.username) <= 8)
        assert(len(self.user_pass) <= 30)
        assert(len(self.nick) <= 8)
        assert(len(self.email) <= 26)

    # Used for creating keys in the key-value store database for this session.
    def db_key(self, sub_key):
        return f"{self.irc_server}/{sub_key}"
    
    # The session stores a list of channels its registered as a set.
    async def db_load_chan_list(self):
        chan_list = set()
        if self.db is not None:
            key_name = self.db_key("chan_list")
            chan_list = await self.db.get(key_name, chan_list)

        return chan_list
    
    """
    Since IRC servers may go up and down or name registrations may fail
    its important to track the state of name registrations. This is
    a higher level function that works on the level of names rather
    than channels. It determines the status of this threshold name
    registration as it relates to this session.
    """
    async def db_is_name_registered(self, name, tld, pw=""):
        dns = {
            "name": name,
            "tld": tld,
            "pw": pw
        }

        # Get list of unique channels. 
        chan_list = await self.db_load_chan_list()
        for chan_name in chan_list:
            # Each chan name indexes meta data related to a name registration.
            key_name = self.db_key(f"chan/{chan_name}")
            chan_meta = await self.db.get(key_name, {})
            if "dns" not in chan_meta:
                continue
            
            # If the meta data is not the name we're looking for skip it.
            if dns != chan_meta["dns"]:
                continue

            # If the name was registered already return true.
            if chan_meta["status"] == IRC_REGISTER_SUCCESS:
                return True

        # Otherwise return false.
        return False
    
    # Connect to a particular IRC server using interface i with AF [0].
    # Also handles login / registration.
    async def start(self, i, timeout=60):
        async def do_start():
            # Choose a supported AF.
            af = i.supported()[0]

            # Destination of IRC server to connect to.
            # For simplicity all IRC servers support v4 and v6.
            dest = await Address(
                self.server_info["ip"][af],
                6697,
                i.route(af)
            )

            # Connect to IRC server.
            route = await i.route().bind()
            self.con = await pipe_open(
                TCP,
                route,
                dest,
                msg_cb=self.msg_cb,
                conf=IRC_CONF
            )

            # Tell server user for the session.
            await self.con.send(
                IRCMsg(
                    cmd="USER",
                    param=f"{self.username} * {self.irc_server}",
                    suffix="*"
                ).pack()
            )

            await self.con.send(
                IRCMsg(cmd="NICK", param=self.nick).pack()
            )


            # Wait for message of the day.
            await asyncio.wait_for(
                self.get_motd, 30
            )
            log("get motd done")

            # Set last started time.
            if self.db is not None:
                last_started_key = self.db_key("last_started")
                await self.db.put(last_started_key, time.time())

            # Trigger register if needed.
            await self.register_user()
            log("register user done")

            # Wait for login success.
            # Some servers scan for open proxies for a while.
            await asyncio.wait_for(
                self.login_status, 15
            )
            log("get login status done.")

            # Load pre-existing owned channels.
            #await self.load_owned_chans()

            # Save user details if needed.
            if self.db is not None:
                nick_key = self.db_key("nick")
                await self.db.put(nick_key, {
                    "domain": self.irc_server,
                    "nick": self.nick,
                    "username": self.username,
                    "user_pass": self.user_pass,
                    "email": self.email,
                    "last_refresh": time.time()
                })

            self.started.set_result(True)
            return self
        
        return await asyncio.wait_for(
            do_start(),
            timeout=timeout
        )

    
    """
    Name -> 20 byte hash160
    (2 ** 160) - 1 yields 
    1461501637330902918203684832716283019655932542975

    IRC chan limit: 31 bytes (1 remaining for #)
    where only a-z0-9 are guaranteed
    that leaves
    ((26 + 10) ** 31) possibilities or
    1759452407304813269615619081855885739163790606335

    so all hash160 values can be stored with this encoding.
    It also allows names to have special characters and be
    longer than 20 bytes (albeit at the cost of collisions.)
    """
    async def get_irc_chan_name(self, name, tld, pw="", executor=None):
        # Domain names are unique per server.
        msg = to_b(f"{name} {tld} {pw} {self.irc_server}")
        if msg in self.chan_name_hashes:
            return self.chan_name_hashes[msg]
        
        if executor is None:
            h, time_lock = f_chan_pow(msg)
        else:
            loop = asyncio.get_event_loop()
            h, time_lock = await loop.run_in_executor(
                executor,
                f_chan_pow,
                (msg)
            )

        self.chan_name_hashes[msg] = h
        self.chan_time_locks[h] = time_lock
        return h
    
    # Close this session gracefully (send IRC QUIT command.)
    # This prevents many timed out sessions on the IRC server.
    async def close(self):
        if self.con is not None:
            await self.con.send(IRCMsg(cmd="QUIT").pack())
            await self.con.close()

    # Look up the value stored in a channel topic.
    async def get_chan_topic(self, chan_name):
        # Setup topic future.
        self.chan_topics[chan_name] = asyncio.Future()

        # First join the channel.
        await self.con.send(
            IRCMsg(
                cmd="JOIN",
                param=f"{chan_name}",
            ).pack()
        )

        # Request the channels topic.
        await self.con.send(
            IRCMsg(
                cmd="TOPIC",
                param=f"{chan_name}",
            ).pack()
        )

        # Wait for channel topic.
        return await self.chan_topics[chan_name]
    
    # Send commands to register the user from the seed in this class.
    async def register_user(self):
        # Build register command.
        suffix = f"REGISTER"
        for param in self.server_info["nick_serv"]:
            if param == "password":
                suffix += f" {self.user_pass}"
            if param == "email":
                suffix += f" {self.email}"

        # Attempt to register the nick at the server.
        # See if the server requires confirmation.
        await self.con.send(
            IRCMsg(
                cmd="PRIVMSG",
                param="NickServ",
                suffix=suffix
            ).pack()
        )

        return self

    # Returns channel owner if registered or False.
    async def is_chan_registered(self, chan_name):
        self.chan_infos[chan_name] = asyncio.Future()
        await self.con.send(
            IRCMsg(
                cmd="PRIVMSG",
                param="ChanServ",
                suffix=f"INFO {chan_name}"
            ).pack()
        )

        return await asyncio.wait_for(
            self.chan_infos[chan_name],
            10
        )
    
    # Register a new channel under our scheme.
    # Many modes are set to minimize abuse.
    async def register_chan(self, chan_name, chan_desc="desc"):
        if chan_name in self.chan_registered:
            return True
    
        # Used to indicate success.
        self.chan_registered[chan_name] = asyncio.Future()
        irc_chan = IRCChan(chan_name, self)

        # Join channel.
        await self.con.send(
            IRCMsg(
                cmd="JOIN",
                param=f"{chan_name}",
            ).pack()
        )

        # Build register command syntax.
        suffix = f"REGISTER {chan_name}"
        if "chan_serv" in self.server_info:
            for param in self.server_info["chan_serv"]:
                if param == "password":
                    suffix += f" {irc_chan.chan_pass}"
                if param == "description":
                    suffix += f" {chan_desc}"

        # register channel.
        await self.con.send(
            IRCMsg(
                cmd="PRIVMSG",
                param="ChanServ",
                suffix=suffix
            ).pack()
        )

        # If the users nickname expires their channels will be dropped.
        # Allow such channels to be recovered.
        await asyncio.sleep(0.1)
        successor = self.server_info["successor"]
        await self.con.send(
            IRCMsg(
                cmd="PRIVMSG",
                param="ChanServ",
                suffix=f"SET {chan_name} SUCCESSOR {successor}"
            ).pack()
        )

        # Attempt to enable topic retention.
        # So the channel topic remains after the last user leaves.
        await asyncio.sleep(0.1)
        self.chans[chan_name] = irc_chan
        await self.con.send(
            IRCMsg(
                cmd="PRIVMSG",
                param="ChanServ",
                suffix=f"SET {chan_name} KEEPTOPIC ON"
            ).pack()
        )

        """
        Todo: The servers do tell you what nodes they support in
        the join message so you could subtract what modes they
        will error on and send it all as one message. But that's
        a lot of work for such an optimization.
        """
        # +s make channel secret so it doesn't show in list
        # We don't want to spam list with a bunch of non-chat channels.
        for mode in "s":
            # Avoid flooding server.
            await asyncio.sleep(0.1)

            # Set the mode.
            await self.con.send(
                IRCMsg(
                    cmd="MODE",
                    param=f"{chan_name} +{mode}",
                ).pack()
            )

        # -k remove password
        # -i no invite only
        for mode in "ki":
            # Avoid flooding the server.
            await asyncio.sleep(0.1)

            # IRC applies modes all or nothing.
            # Try to get the 'most open' channel.
            await self.con.send(
                IRCMsg(
                    cmd="MODE",
                    param=f"{chan_name} -{mode}",
                ).pack()
            )

        # Record list of channels that belongs to this session.
        if self.db is not None:
            # Chan list update.
            chans_key = self.db_key("chan_list")
            chan_list = await self.db.get(chans_key, [])
            chan_list.add(chan_name)
            await self.db.put(chans_key, chan_list)

            # Chan meta data.
            await self.db.put(self.last_started, {
                "last_refresh": time.time()
            })

        return True
    
    """
    Called asynchronously when a new message arrives on the pipe.
    These messages may be partial so they're appended to an existing
    buffer and a function processes them for validity.
    Then another function that implements a portion of the IRC
    protocol is called on each valid IRC message with any replies
    forwarded back to the server pipe.
    """
    async def msg_cb(self, msg, client_tup, pipe):
        log(msg)

        try:
            # Keep a buffer of potential protocol messages.
            # These may be partial in the case of TCP.
            self.recv_buf += to_s(msg)
            msgs, new_buf = irc_extract_msgs(self.recv_buf)
            self.recv_buf = new_buf

            # Disable chan success in some cases.
            skip_register = False
            for msg in msgs:
                for chan_fail in IRC_NEG_CHAN_MSGS:
                    if chan_fail in msg.suffix:
                        skip_register = True
                        break

            # Loop over the IRC protocol messages.
            # Process the minimal functions we understand.
            for msg in msgs:
                resp = irc_proto(self, msg)
                if resp is not None:
                    await pipe.send(resp.pack())
    
        except Exception:
            log_exception()

"""
The main 'manager' that handles threshold consensus of name registrations
and lookups using IRC. Has a list of 'sessions' representing different
IRC servers and some simple limits that determine success.
"""
class IRCDNS():
    def __init__(self, i, seed, servers, executor=None, clsChan=IRCChan, clsSess=IRCSession, do_shuffle=True):
        self.i = i
        self.seed = seed
        self.sessions = {}
        self.p_sessions_next = 0

        """
        The classes to use to create channels and sessions are dynamic.
        The reason for this is it allows for mock classes to be dynamically
        passed to IRCDNS for the purposes of testing core logic within
        this class. Otherwise the above two classes are used.
        """
        self.clsChan = clsChan
        self.clsSess = clsSess

        """
        Executors represent processpoolexecutors that allow for the
        verifiable delay function to be 'saved up' in parallel and
        used to register multiple channels concurrently. This
        makes things faster and prevents race conditions.
        """
        self.executor = executor
        if self.executor is None:
            # We will handle executor cleanup.
            self.close_executor = True
            self.executor = ProcessPoolExecutor()
        else:
            # Caller has to clean up their executors.
            self.close_executor = False

        """
        To help with load balancing -- servers may be shuffled.
        This leads to non-determinism which makes testing harder so
        it can be disabled if needed.
        """
        if do_shuffle:
            random.shuffle(servers)

        self.servers = servers

        """
        The start_n function modifies a pointer called p_sessions_next
        that points to the offset in the sessions dictionary to store the
        next session at. If multiple coroutines call start_n and are interrupted
        the pointers value may be malformed leading to sessions being overwritten.
        """
        self.start_lock = asyncio.Lock()

        """
        Key information about names is stored in a simple SQlite database using
        a simple key-value approach. The database location is determined by
        the seed and prefix passed to the class.
        """
        self.db = SqliteKVS(
            os.path.join(
                os.path.expanduser("~"),
                "ircdns_" + sha3_256(IRC_PREFIX + b"db" + IRC_SALT + self.seed) + ".sqlite"
            )
        )

    async def start(self):
        await self.db.start()

        # Save seed to DB.
        await self.db.put("seed", self.seed)

        return self

    # Don't include servers older than 48 hours.
    """
    The idea of this function is to allow for the set of what are
    considered 'valid servers' to be dynamically adjusted. If such
    a set is fixed and they determine consensus for registration success
    and lookups then servers that go offline forever may break the module.
    Similarly, servers that disable registrations would also qualify.
    """
    async def get_server_len(self):
        # No sessions to base exclusion on.
        if not len(self.sessions):
            return len(self.servers)
        
        # Exclude servers that have been offline a long time.
        count = 0; start = time.time()
        for n in range(0, len(self.servers)):
            last_started_key = f"{self.servers[n]['domain']}/last_started"
            last_started = await self.db.get(last_started_key, start)
            duration = start - last_started
            if duration < IRC_FRESHNESS:
                count += 1

        return count

    """
    These are all simple math functions to determine how many servers
    can fail and calls still succeed. I've chosen that up to 40% can fail.
    This number results in few servers in a list of 5 and seems to feel right.
    """
    async def get_register_failure_max(self):
        return int(0.4 * await self.get_server_len())
    
    async def get_register_success_min(self):
        return await self.get_server_len() - await self.get_register_failure_max()
    
    async def get_register_success_max(self):
        return await self.get_server_len()
    
    async def get_lookup_success_min(self):
        return await self.get_register_failure_max() + 1
    
    """
    When calling 'start' on a session there is a future set at the very end
    of login / registration / get message of the day to indicate success.
    This function counts success by looking at those set futures.
    """
    def count_started_sessions(self):
        # Count existing 
        success_no = 0
        for n in range(0, self.p_sessions_next):
            if self.sessions[n].started.done():
                success_no += 1

        return success_no

    # Close all open pipes to IRC servers and close the database.
    async def close(self):
        tasks = []
        for i in range(0, self.p_sessions_next):
            task = self.sessions[i].close()
            tasks.append(task)

        await asyncio.gather(*tasks)
        await self.db.close()
        if self.close_executor:
            if self.executor is not None:
                self.executor.shutdown(wait=False)
                self.executor = None

    # Start n new sessions.
    async def start_n(self, n):
        await self.start_lock.acquire()
        try:
            assert(self.p_sessions_next <= len(self.servers))

            # Are there enough unstarted servers to try?
            tasks = []; p = self.p_sessions_next
            if (p + n) > len(self.servers):
                raise Exception("No IRC servers left to try.")

            # Create sessions from any already started.
            for j in range(p, p + n):
                assert(j not in self.sessions)
                self.sessions[j] = self.clsSess(
                    self.servers[j],
                    seed=self.seed,
                    db=self.db,
                    offset=j,
                )

                tasks.append(
                    async_wrap_errors(
                        self.sessions[j].start(self.i)
                    )
                )
                self.p_sessions_next += 1

            # Start them all at once if needed.
            if len(tasks):
                await asyncio.gather(*tasks)

            # Determine sessions that worked.
            success_no = 0
            failure_offsets = []
            for j in range(0, self.p_sessions_next):
                if self.sessions[j].started.done():
                    success_no += 1
                else:
                    failure_offsets.append(j)
            return success_no, failure_offsets
        finally:
            self.start_lock.release()

    """
    The verifiable delay function (Argon2) is slow and its used
    to calculate what a channels name should be. To speed things up
    a simple table of channel names is saved from higher-level names
    and returned instantly when requesting the channel name.
    This greatly simplifies the existing code. You can see that
    this occurs in parallel using process executors (different CPU cores.)
    """
    async def pre_cache(self, name, tld, pw=""):
        tasks = []
        for n in range(0, self.p_sessions_next):
            task = async_wrap_errors(
                self.sessions[n].get_irc_chan_name(
                    name,
                    tld,
                    pw,
                    self.executor
                )
            )

            tasks.append(task)

        if len(tasks):
            await asyncio.gather(*tasks)
    
    """
    This is a massive function that has many features.
        - It ensures names are only registered if there's enough sessions
        where the names are available to satisfy threshold requirements
        (hence you could call this 'atomic.')
        - It is able to handle partial registrations that result from past
        calls (by restarting sessions that may have not been available.)
        - It is able to handle differences between name availability and
        session failures (to maximize success.)
        - It handles state by recording registrations across servers in
        a simple local database using a key-value store.
        - It enables successive calls to be done on the same name and takes
        into account names that you already know (so that the call can easily
        be used to attempt to register names on sessions that weren't online.)

    It's the most important function in this module.
    """
    async def name_register(self, name, tld, pw=""):
        # Ensure the session pointer hasn't overflowed.
        assert(self.p_sessions_next <= len(self.servers))

        # Open the bare minimum number of sessions
        # for a successful consensus result.
        success_no, failure_offsets = await self.start_n(
            len(self.servers) - min(
                self.p_sessions_next,
                len(self.servers)
            )
        )
        if success_no < await self.get_register_success_min():
            raise Exception("Insufficient sessions.")
        
        # Pre-cache chan name hashes.
        await self.pre_cache(name, tld, pw)
        
        # Check if name available on servers.
        tasks = []
        for n in range(0, self.p_sessions_next):
            async def is_name_available(n):
                # Simulate name being available if we already own it.
                # This will allow the code to pass and be reused.
                already_reg = await (self.sessions[n].db_is_name_registered(name, tld, pw))
                if already_reg:
                    return True

                # Convert name to server-specific hash.
                chan_name = await self.sessions[n].get_irc_chan_name(
                    name,
                    tld,
                    pw,
                    self.executor
                )

                # Check if resulting channel is registered.
                ret = await self.sessions[n].is_chan_registered(chan_name)

                """
                Return that a name is available if we already own it.
                This will trick the algorithm into succeeding without
                special modifications and makes it 'smarter.'
                Success on register isn't checked for anyway. It's
                just assumed that it will work if names are available.
                """
                if ret == self.sessions[n].nick:
                    return True
                
                # Channel registered to someone else.
                if ret != False:
                    return False
                else:
                    return True
            
            # Skip sessions that aren't connected.
            if self.sessions[n].started.done():
                tasks.append(
                    async_wrap_errors(
                        is_name_available(n)
                    )
                )

        # Run availability checks concurrently.
        results = await asyncio.gather(*tasks)
        
        # Check if there's enough names available.
        available_count = 0
        for is_chan_available in results:
            if is_name_available is not None and is_chan_available:
                available_count += 1
        if available_count < await self.get_register_success_min():
            raise Exception("Not enough names available.")
        
        # Register name.
        tasks = []
        for n in range(0, self.p_sessions_next):
            # Helper function to register a name for a session.
            async def do_register(n):
                # Attempt to start session if not started.
                # Useful for the register_factory code.
                if not self.sessions[n].started.done():
                    try:
                        await self.sessions[n].start(self.i)
                    except:
                        return [
                            n,
                            False
                        ]

                # Name already registered so skip.
                already_reg = await (self.sessions[n].db_is_name_registered(name, tld, pw))
                if already_reg:
                    return [
                        n,
                        self.sessions[n].nick
                    ]

                # Convert name to server-specific hash.
                chan_name = await self.sessions[n].get_irc_chan_name(
                    name,
                    tld,
                    pw,
                    self.executor
                )

                # Register name on that server.
                await self.sessions[n].register_chan(chan_name)
                await asyncio.sleep(0.1)

                # Return owner of channel.
                return [
                    n,
                    await self.sessions[n].is_chan_registered(chan_name)
                ]

            tasks.append(
                async_wrap_errors(
                    do_register(n)
                )
            )

        # Do registration tasks concurrently.
        results = await asyncio.gather(*tasks)
        await asyncio.sleep(0.1)

        # Name portion to store in record.
        dns_meta = {
            "name": name,
            "tld": tld,
            "pw": pw
        }

        # Get owners for channels to see success.
        records = []; error_no = IRC_REGISTER_SUCCESS
        for n in range(0, self.p_sessions_next):
            # Default to failure.
            reg_status = IRC_REGISTER_FAILURE

            # The server couldn't be connected.
            if n in failure_offsets:
                reg_status = IRC_START_FAILURE

            # Check channel owner after registration.
            # Results need to exist for success to be possible.
            for r in results:
                # Exception occurred.
                if r is None:
                    continue

                # If the session offset is this and
                # the session nick matches the channel nick.
                if n == r[0] and r[1] == self.sessions[n].nick:
                    reg_status = IRC_REGISTER_SUCCESS
                    break

            # Set overall status.
            if reg_status != IRC_REGISTER_SUCCESS:
                error_no = IRC_REGISTER_PARTIAL

            # Get the chan name.
            # Calling this here also ensures time_lock exists.
            chan_name = await async_wrap_errors(
                self.sessions[n].get_irc_chan_name(
                    name,
                    tld,
                    pw,
                    self.executor
                )
            )
            if chan_name is None:
                continue

            # Store chan info here.
            chan_info_key = self.sessions[n].db_key(f"chan/{chan_name}")

            # Get the time-lock field.
            time_lock = self.sessions[n].chan_time_locks[chan_name]

            # Get list of channels for this server.
            chan_list_key = self.sessions[n].db_key("chan_list")
            chan_list = await self.db.get(chan_list_key, set())
            chan_list.add(chan_name)
            await self.db.put(chan_list_key, chan_list)

            # Get failure count for registration.
            failure_count = 0
            old_record = await self.db.get(chan_info_key, {})
            if "failure_count" in old_record:
                failure_count = old_record["failure_count"]
            if reg_status != IRC_REGISTER_SUCCESS:
                failure_count += 1

            # Record to store about name in session.
            record = {
                "domain": self.sessions[n].irc_server,
                "dns": dns_meta,
                "chan_name": chan_name,
                "time_lock": time_lock,
                "status": reg_status,
                "failure_count": failure_count,
                "last_refresh": time.time()
            }

            # Record the changes.
            await self.db.put(chan_info_key, record)
            records.append(record)

        # Store name in DB.
        if self.db is not None:
            dns_names = await self.db.get("names", set())
            dns_names.add(f"{name}//{tld}//{pw}")
            await self.db.put("names", dns_names)

        return records, error_no
    
    """
    This function handles storing a value at a given name across
    a threshold of IRC servers. It will ensure that the name has a unique
    'identity' (through an ECDSA key); that the identity is encrypted;
    that the value is encrypted; and that its hard to determine what the name
    is in order to improve privacy. Timestamps are also part of the
    structure and the lookup function prioritizes freshness where possible.
    """
    async def store_value(self, value, name, tld, pw=""):
        async def helper(n):
            # Select session to use.
            session = self.sessions[n]
            if not session.started.done():
                return

            topic, chan = await f_pack_topic(
                value,
                name,
                tld,
                pw,
                session,
                self.clsChan,
                self.executor
            )

            # Reference function to set topic.
            await chan.set_topic(topic)

        # Open as many sessions as possible.
        await self.start_n(
            len(self.servers) - self.p_sessions_next
        )

        # Pre-cache name hashes.
        await self.pre_cache(name, tld, pw)

        # Build tasks to set chan topics.
        tasks = []
        for n in range(0, len(self.servers)):
            tasks.append(
                async_wrap_errors(
                    helper(n)
                )
            )

        # Execute tasks to update topics.
        await asyncio.gather(*tasks)

    # Ensures that a certain number of sessions are started.
    async def acquire_at_least(self, target):
        # Count existing 
        success_no = self.count_started_sessions()

        # Not enough sessions open to satisfy target.
        # Open needed sessions.
        if success_no < target:
            still_needed = target - success_no

            # Progress through server list until target is met
            # otherwise the end is encountered and an exception is thrown.
            while still_needed:
                success_no, _ = await self.start_n(still_needed)
                still_needed -= success_no

    """
    For a lookup request to succeed a certain number of results need to
    be gathered - enough so that it is beyond doubt that the gathered
    results represent a name stored on a subset of servers that are large
    enough to count as a valid registration (this prevents attacks.)
    """
    async def n_more_or_best(self, results):
        if not len(results):
            return await self.get_lookup_success_min()
        
        table = {}
        for r in results:
            if r is None:
                continue

            for r_id in r["ids"]:
                if r_id not in table:
                    table[r_id] = [r]
                else:
                    table[r_id].append(r)

        highest = []
        for r_id in table:
            if len(highest) < len(table[r_id]):
                highest = table[r_id]

        if len(highest) >= await self.get_lookup_success_min():
            freshest = highest[0]
            for result in highest:
                if result["time"] > freshest["time"]:
                    freshest = result

            return freshest
        else:
            return await self.get_lookup_success_min() - len(highest)
        
    """
    This is a bulk function that handles making individual lookup
    requests to each IRC session for a name and decoding and returned values.
    The above function will still need to succeed to choose the best result.
    """
    async def n_name_lookups(self, n, start_p, name, tld, pw=""):
        async def helper(self, session_offset, chan_name):
            session = self.sessions[session_offset]
            topic = await asyncio.wait_for(
                session.get_chan_topic(chan_name),
                10
            )
            
            return f_unpack_topic(
                chan_name,
                topic,
                session
            )

        tasks = []; p = start_p
        while len(tasks) < n:
            if p >= len(self.servers):
                raise Exception("Exceeded irc sessions.")

            if not self.sessions[p].started.done():
                p += 1
                continue

            # Get chan name.
            chan_name = await self.sessions[p].get_irc_chan_name(
                name,
                tld,
                pw,
                self.executor
            )

            # Get name or throw an error.
            tasks.append(
                async_wrap_errors(
                    helper(self, p, chan_name),
                    5
                )
            )

            p += 1

        if len(tasks):
            return await asyncio.gather(*tasks), p
        else:
            return [], p
        
    """
    This is a high-level function used to translate a name into a value
    that is stored across multiple IRC servers for integrity protection.
    It will decrypt the value, handle identity verification, prioritize
    freshness, and ensure there's enough results to constitute a valid name.
    """
    async def name_lookup(self, name, tld, pw=""):  
        # Pre-cache name hashes.
        await self.pre_cache(name, tld, pw)

        # Open the bare minimum number of sessions
        # for a successful consensus result.
        await self.acquire_at_least(
            await self.get_lookup_success_min()
        )

        # Get minimum number of chan topics.
        results = []; start_p = 0
        names_required = await self.n_more_or_best(results)
        while isinstance(names_required, int):
            more, start_p = await self.n_name_lookups(
                names_required,
                start_p,
                name,
                tld,
                pw
            )

            results += more
            names_required = await self.n_more_or_best(results)

        # Return freshest result.
        assert(len(results) >= await self.get_lookup_success_min())
        if isinstance(names_required, dict):
            freshest = names_required
            return freshest
        
"""
On IRC networks nicknames that don't login for a certain period will expire
and their associated channels will be dropped. Likewise, channels with
no activity will expire. For this solution to be practical there needs to
be a way to check if any nicks or channels are close to expiry and refresh
them. Currently this is a simple call to refresh on an IRCDNS object. I've
also made the refresher call register on names that were only partially
registered (up to a count of 5) in order to be more fault-tolerant.
"""
class IRCRefresher():
    def __init__(self, manager):
        self.manager = manager

    async def placeholder(self):
        return

    async def refresh_chan(self, chan_name, session):
        # Session not started.
        if not session.started.done():
            await session.start(self.manager.i)

        # Join channel.
        await session.con.send(
            IRCMsg(
                cmd="JOIN",
                param=f"{chan_name}",
            ).pack()
        )

        # Send a basic refresh message.
        await session.con.send(
            IRCMsg(
                cmd="PRIVMSG",
                param=f"{chan_name}",
                suffix=IRC_CHAN_REFRESH
            ).pack()
        )

    async def refresh_nick(self, session):
        # Session not started.
        if not session.started.done():
            await session.start(self.manager.i)

        # Join testing channel.
        chan_name = f'{session.server_info["test_chan"]}'
        await session.con.send(
            IRCMsg(
                cmd="JOIN",
                param=chan_name,
            ).pack()
        )

        # Send a basic refresh message.
        await session.con.send(
            IRCMsg(
                cmd="PRIVMSG",
                param=chan_name,
                suffix=IRC_NICK_REFRESH
            ).pack()
        )

    async def refresher(self, ignore_failure=False):
        tasks = []
        db = self.manager.db

        # Return coroutine that does nothing if no refresh needed.
        # Otherwise returns the coroutine that does the refresh.
        async def check_last_refresh(sub_key, expiry_days, expiry_func):
            # Extract meta data.
            key_name = session.db_key(sub_key)
            info = await db.get(key_name, {})
            if "last_refresh" not in info:
                return self.placeholder()

            # Calculate the expiry time in seconds.
            day_secs = 24 * 60 * 60
            expiry_secs = expiry_days * day_secs

            # Refresh five days before expiry.
            expiry_secs -= 5 * day_secs
            assert(expiry_secs > 0)
            duration = max(time.time() - info["last_refresh"], 0)
            if duration >= expiry_secs:
                # Update refresh timer.
                info["last_refresh"] = time.time()
                await db.put(key_name, info)

                # Factory that returns coroutine to await on.
                return async_wrap_errors(
                    expiry_func()
                )
            
            # So results can be passed to gather.
            return self.placeholder()

        # Loop over sessions for anything that needs refreshing.
        register_list = []
        for n in range(0, self.manager.p_sessions_next):
            # Select valid session.
            session = self.manager.sessions[n]
            
            # Loop over all channels registered to session.
            chan_list = await session.db_load_chan_list()
            for chan_name in chan_list:
                # Check the channel info for expiry.
                chan_info_key = session.db_key(f"chan/{chan_name}")
                chan_info = await session.db.get(chan_info_key, {})
                expiry_days = session.server_info["chan_expiry"]
                tasks.append(
                    await check_last_refresh(
                        f"chan/{chan_name}",
                        expiry_days,
                        lambda: self.refresh_chan(chan_name, session)
                    )
                )

                # Sanity checks for the next part.
                # Only attempt to re-register up to 3 times.
                # Don't want to DoS server.
                if "status" not in chan_info:
                    continue 
                if "dns" not in chan_info:
                    continue

                if not ignore_failure:
                    if "failure_count" not in chan_info:
                        chan_info["failure_count"] = 0
                    if chan_info["failure_count"] >= 3:
                        continue

                # If the chan info status was start failure
                # try to call register on the name again.
                if chan_info["status"] == IRC_START_FAILURE:
                    # Register handles retry for all servers.
                    if chan_info["dns"] not in register_list:
                        tasks.append(
                            self.manager.name_register(
                                name=chan_info["dns"]["name"],
                                tld=chan_info["dns"]["tld"],
                                pw=chan_info["dns"]["pw"]
                            )
                        )
                        register_list.append(chan_info["dns"])

            # See if nick needs to be refreshed.
            tasks.append(
                await check_last_refresh(
                    "nick",
                    session.server_info["nick_expiry"],
                    lambda: self.refresh_nick(session)
                )
            )
            
        # Execute refresh tasks concurrently.
        if len(tasks):
            await asyncio.gather(*tasks)

if __name__ == '__main__':
    pass



"""
This code is based on the original implementation of TURN as a TCP proxy for SOCKs requests which is hosted here: https://github.com/trichimtrich/turnproxy

My observation is that TURNs connect method for TCP assumes that the destination is reachable. In a P2P context this is useless because peers need to use TURN as a way for two hosts to contact each other when all other methods fail. If a host can be contacted directly then there is no reason to use a TURN server. Hence TURNs TCP model is useless (no -- it's not setup for TCP hole punching either.)

Instead, I focus on using UDP as a last resort (UDP has packet loss and no ordering which is really frigging annoying for writing software.) UDP has the advantage of automatically openning a hole in a users NAT / router when sending an outbound UDP packet. This means that clients who connect to a TURN server will be reachable, unlike the TCP connect method.

The changes I've implemented:
- UDP instead of TCP
- Support multiple network interfaces
- IPv6 support
- Multiple client sessions simultaneously
- Full compatibility with the Pipe object

Note to self:
Don't try use TURNs TCP connection mode again. Even using an edge-case where you can get TURN to connect to its own relay addresses (in Coturn) doesn't work because Coturn limits active connections to 1 per IP. Therefore it would only support a one-way channel to one peer. Very much useless. I have found no way around TURNs limitations for TCP and think it was poorly designed. TURN is the worst protocol I have ever implemented so this does not surprise me.

https://datatracker.ietf.org/doc/html/draft-ietf-behave-turn-tcp-07

TODO: Future feature = implement shared secrets 
https://datatracker.ietf.org/doc/html/draft-rosenberg-midcom-turn-08#page-9

matrix.org seems to use them over static credentials
"""

import asyncio
from struct import pack
from .address import *
from .interface import *
from .turn_defs import *
from .turn_process import *
from .base_stream import *

# Main class for handling TURN sessions with a server.
class TURNClient(BaseProto):
    def __init__(
        self,
        route: any,
        turn_addr: Address,
        turn_user: bytes = None, # turn username
        turn_pw: bytes = None, # turn password
        turn_realm: bytes = None,
        msg_cb: any = None,
        conf: any = NET_CONF
    ):
        # Can received relay messages have a blank header?
        self.blank_rudp_headers = False

        # Remote address for the TURN server.
        # Username and password are optional.
        self.requires_auth = True
        self.turn_addr = turn_addr
        self.turn_user = turn_user
        self.turn_pw = turn_pw
        self.msg_cb = msg_cb

        # Set from attributes in replies.
        self.realm = turn_realm
        if turn_addr.host_type == HOST_TYPE_DOMAIN:
            self.realm = turn_addr.host
        self.key = None
        self.nonce = None

        # The main UDP endpoint used to talk to the client.
        # route = NIC bind details to use for the pipe.
        self.turn_pipe = None
        self.route = route
        self.conf = conf

        # Special attribute set to indicate expiry time of an allocation.
        self.lifetime = TURN_REFRESH_EXPIRY

        # Our own peer address.
        self.mapped = []
        self.relay_tup = None

        # The initial session is associated with a random TXID.
        # Replies specify that TXID so the same session can be identified.
        self.txid = b""
        self.con_id = None

        # Event set when protocol completed and chan messages can be sent.
        self.processing_loop_task = None
        #self.allocate_refresher_task = None

        # The protocol client uses a state machine.
        # Each state has a set duration for it to be completed in.
        self.state_timestamp = timestamp()
        self.state = TURN_NOT_STARTED
        self.peers = {}
        self.msgs = {}
        self.tasks = []

        # Event loop reference.
        loop = asyncio.get_event_loop()
        if self.conf["loop"] is not None:
            loop = self.conf["loop"]()

        # Futures to return from start.
        self.turn_client_stopped = asyncio.Event()
        self.client_tup_future = loop.create_future()
        self.relay_tup_future = loop.create_future()
        self.auth_event = asyncio.Event()
        self.relay_event = asyncio.Event()
        self.node_events = {} # by node_id

    def get_turn_server(self, af=None):
        return {
            "host": self.turn_addr.host,
            "port": self.turn_addr.port,
            "afs": [af],
            "user": self.turn_user,
            "pass": self.turn_pw,
            "realm": self.realm
        }
    
    def get_relay_tup(self, peer_ip):
        if peer_ip in self.peers:
            return self.peers[peer_ip]
        else:
            return None

    def toggle_blank_rudp_headers(self, val):
        self.blank_rudp_headers = val

    # Make this whole clas look like a 'pipe' object.
    def super_init(self,  transport, sock, route, conf=NET_CONF):
        super().__init__(sock=sock, route=route, conf=conf)
        self.connection_made(transport)
        self.stream.set_handle(transport, client_tup=None)

    # Start the TURN client.
    async def start(self, n=0):
        # Set and validate peer address.
        log("> Turn starting client.")
        if self.turn_user is None and self.turn_pw is None:
            self.auth_event.set()

        # Connect to TURN server over UDP.
        af = self.turn_addr.af
        self.turn_pipe = await pipe_open(
            route=self.route,
            proto=UDP,
            dest=self.turn_addr
        )
        log(f"> Turn socket = {self.turn_pipe.sock}")

        # If con was unncessessful raise exception.
        if self.turn_pipe is None:
            raise Exception("Unable to connect to TURN host. This may mean the server is no longer working. Normally TURN is not a public service.")

        # Subscribe to all messages.
        self.turn_pipe.subscribe(SUB_ALL)

        # Make this entire class shadow the pipe above.
        self.super_init(
            transport=self.turn_pipe.transport,
            sock=self.turn_pipe.sock,
            route=self.route, 
            conf=self.conf
        )

        # Start processing UDP replies.
        self.processing_loop_task = asyncio.create_task(
            async_wrap_errors(
                process_replies(self)
            )
        )
        #self.tasks.append(self.processing_loop_task)
        
        # Add any message handlers.
        if self.msg_cb is not None:
            self.add_msg_cb(self.msg_cb)

        # expect 'Unauthorized'.
        await async_retry(lambda: self.allocate_relay(sign=False), count=5)

        # Wait for client to be ready.
        await self.auth_event.wait() # Authentication success.
        await self.relay_event.wait() # Our relay address available.

        # Return our relay tup.
        relay_tup = await self.relay_tup_future
        client_tup = await self.client_tup_future

        # White list ourselves.
        #await self.accept_peer(client_tup, relay_tup)

        # Refresh allocations.
        async def refresher():
            while 1:
                await asyncio.sleep(TURN_REFRESH_EXPIRY - 60)
                try:
                    await async_retry(
                        lambda: self.refresh_allocation(),
                        count=5,
                        timeout=5
                    )
                except Exception:
                    try:
                        await self.reconnect(n=1)
                    except Exception:
                        log_exception()
                        continue

        # First run of this function.
        if not n:
            self.allocate_refresher_task = asyncio.create_task(
                async_wrap_errors(
                    refresher()
                )
            )
            self.tasks.append(self.allocate_refresher_task)

        return relay_tup

    def __await__(self):
        return self.start().__await__()

    async def reconnect(self, n=0):
        await self.close()

        # Overwrite current state.
        self.__init__(
            route=self.route,
            turn_addr=self.turn_addr,
            turn_user=self.turn_user,
            turn_pw=self.turn_pw,
            turn_realm=self.realm,
            msg_cb=self.msg_cb,
            conf=self.conf
        )

        # Try start it again.
        await self.start(n)

    # Changes the protocol state machine.
    def set_state(self, state):
        log("> Turn moving state from %s to %s." % (self.state, state))
        self.state = state

    def new_node_event(self, node_id):
        self.node_events[to_s(node_id)] = asyncio.Event()

    # Overwrite the BaseProto send method and require ACKs.
    async def send(self, data, dest_tup):
        # Detect invalid self-send.
        if self.relay_tup_future.done():
            relay_tup = await self.relay_tup_future
            if dest_tup == relay_tup:
                raise Exception("Coturn doesn't support self-send.")

        # Make sure the channel is setup before continuing.
        task = asyncio.create_task(
            async_wrap_errors(
                self.stream.ack_send(
                    data,
                    dest_tup
                )
            )
        )
        self.tasks.append(task)

    # Handles writing TURN messages to self.udp_stream.
    # Will write credential and HMAC if a message needs 'signing.'
    async def send_turn_msg(self, msg: TurnMessage, do_sign=False):
        buf, _ = TurnMessage.unpack(msg.encode())
        if self.requires_auth:
            if do_sign and self.key:
                buf.write_credential(self.turn_user, self.realm, self.nonce)
                buf.write_hmac(self.key)

        buf = buf.encode()
        await self.turn_pipe.send(buf, self.turn_addr.tup)

    # Record TURN protocol messages by TXID.
    # Events are triggered on receipt.
    def record_msg(self, msg):
        f = asyncio.Future()
        self.msgs[msg.txn_id] = {
            "status": f,
            "timestamp": timestamp(),
            "msg": msg
        }

        def new_future():
            a_future = asyncio.Future()
            self.msgs[msg.txn_id]["status"] = a_future
            return a_future

        def closure():
            async def retransmit():
                await self.send_turn_msg(msg, do_sign=True)

            return retransmit

        return f, closure(), new_future

    # Create and send an allocation request.
    # Results in a new relay address being allocated for the client.
    async def allocate_relay(self, sign):
        msg = await self.allocate_msg()
        f, retransmit, new_future = self.record_msg(msg)
        await self.send_turn_msg(msg, do_sign=sign)
        return f, retransmit, new_future

    # Create and send a create permission for a peers address.
    # Retry up to 3 times if no response to the packet.
    # Allows a peer to send messages to our relay address.
    async def accept_peer(self, peer_tup, peer_relay_tup):
        #peer_tup = (peer_tup[0], 0)
        peer_ip = peer_tup[0]
        already_accepted = peer_ip in self.peers
        async def handler(peer_tup, peer_relay_tup):
            # Generate message to send.
            msg = await self.white_list_msg(peer_tup)
            self.peers[peer_ip] = peer_relay_tup

            # Send message to turn server.
            f, retransmit, new_future = self.record_msg(msg)
            return f, retransmit, new_future

        # Refresh permissions.
        f = lambda: handler(peer_tup, peer_relay_tup)
        async def refresher():
            while self.state != TURN_ERROR_STOPPED:
                await asyncio.sleep(TURN_REFRESH_EXPIRY - 60)
                await async_retry(f, count=5, timeout=5)
                log("Refresh permission.")

        # Prevent garbage collection.
        if not already_accepted:
            # Allow messages to be queued.
            sub = self.tup_to_sub(peer_relay_tup)
            self.subscribe(sub)

            # White list the peer if needed.
            await async_retry(f, count=5, timeout=5)

            # Start the loop to refresh the permission.
            task = asyncio.create_task(
                async_wrap_errors(
                    refresher()
                )
            )
            self.tasks.append(task)

    # Relay addresses are only valid for a certain 'life time.'
    # This creates and sends a message to refresh the lifetime.
    async def refresh_allocation(self):
        log("> Turn refreshing allocate lifetime.")
        msg = await self.refresh_msg()
        f, retransmit, new_future = self.record_msg(msg)
        return f, retransmit, new_future
        
    # Main step 1 -- allocate a relay address msg.
    async def allocate_msg(self):
        reply = TurnMessage(msg_type=TurnMessageMethod.Allocate)
        reply.write_attr(
            TurnAttribute.RequestedTransport,
            TURN_RPOTOCOL_UDP
        )

        self.txid = reply.txn_id
        return reply
        
    # Main step 2 -- white list a peer to use our relay address msg.
    # Apparently the port number is irrelevant.
    # Permissions are made per IP.
    async def white_list_msg(self, src_tup):        
        # Try write the peer address.
        reply = TurnMessage(msg_type=TurnMessageMethod.CreatePermission)
        try:
            turn_write_peer_addr(
                reply,
                src_tup
            )
        except:
            log_exception()
            return None

        return reply

    # Step 3 - refresh allocation to avoid lifetime timeouts.
    async def refresh_msg(self):
        # 32 bit unsigned int
        reply = TurnMessage(msg_type=TurnMessageMethod.Refresh)
        reply.write_attr(
            TurnAttribute.Lifetime,
            pack("!I", TURN_REFRESH_EXPIRY)
        )

        """
        reply.write_attr(
            TurnAttribute.RequestedTransport,
            TURN_RPOTOCOL_UDP
        )
        """

        # Return reply message.
        #reply.txn_id = self.txid
        return reply

    # Close the client socket and move state to done.
    async def do_cleanup(self):
        if self.turn_pipe is not None:
            await self.turn_pipe.close()

        # Make the main message process loop end.
        self.state = TURN_ERROR_STOPPED

        # Make pending TURN handlers finish.
        if not self.client_tup_future.done():
            self.client_tup_future.set_result((0, 0))

        if not self.relay_tup_future.done():
            self.relay_tup_future.set_result((0, 0))

        # Make any pending send or recv calls finish.
        self.auth_event.set()
        self.relay_event.set()

    async def close(self):
        # Already closed.
        if self.turn_client_stopped.is_set():
            return

        # Close all pipes.
        # Set events as done so all tasks end.
        await self.do_cleanup()

        # Processing loop sets this when done.
        await self.turn_client_stopped.wait()

        # Wait for permission refresher tasks or cancel them.
        await gather_or_cancel(self.tasks, 2)

if __name__ == '__main__': # pragma: no cover
    """
    // If left out, will use openrelay public TURN servers from metered.ca
    see if these servers work?
    turnIceServers: { ... },
    """
    async def test_turn():
        """
        buf = b"ur\x00\t\xd6o'\x04\x9ezp*\x01"
        m = TurnMessage.unpack(buf)[0]
        print(m)
        print(m.eof())
        while not m.eof():
            attr_code, _, attr_data = m.read_attr()
            attr_name = TurnAttribute.get(attr_code)

            print(attr_code)
            print(attr_name)

        return
        """
        interface = await Interface("enp1s0f0").start()
        af = AF_INET
        turn_user=b""
        turn_pw=b""
        route = interface.route(af)
        turn_addr = await Address("", 3478, route).res()


        """
        A faulty network interface will cause hosts with multiple
        interfaces to report non-deterministic results with defaults.
        Thus, its better to manually select an interface for testing
        than to silently fail and wonder what is going wrong.
        This interface uses a preserving type nat so it bypasses the
        issue with coturn reply ports.
        """
        client1 = TURNClient(
            turn_addr=turn_addr,
            turn_user=turn_user,
            turn_pw=turn_pw,
            interface=interface
        )

        client_tup_future, relay_tup_future, in_chan_event = await client1.start()
        client_tup = await client_tup_future
        relay_tup = await relay_tup_future

        """
        reply = TurnMessage(msg_type=TurnMessageMethod.Send, msg_code=TurnMessageCode.Indication)
        reply.write_attr(
            TurnAttribute.Data,
            b"send indication test msg."
        )
        turn_write_peer_addr(reply, client_tup)
        await client1.send_turn_msg(reply, do_sign=True)
        """


        while 1:
            await asyncio.sleep(1)


    async_test(test_turn)





import os
import pathlib
import inspect
import shutil

# Path to where the script is running from.
def get_script_parent():
    filename = inspect.getframeinfo(inspect.currentframe()).filename
    parent = pathlib.Path(filename).resolve().parent
    return os.path.realpath(parent)

# Home dir / p2pd.
def get_p2pd_install_root():
    return os.path.realpath(
        os.path.join(
            os.path.expanduser("~"),
            "p2pd"
        )
    )

# Home dir / p2pd / kvs.sqlite3.
# Used to store ECDSA key pairs that own PDNS names.
def get_kvs_db_install_path(install_root):
    return os.path.realpath(
        os.path.join(
            install_root,
            "kvs.sqlite3"
        )
    )

# Location in p2pd install where the blank KVS db lives.
def get_kvs_db_copy_path(script_parent):
    return os.path.realpath(
        os.path.join(
            script_parent,
            "scripts",
            "kvs_schema.sqlite3"
        )
    )

# Installs P2PD files into home dir.
# The software only needs this for using PDNS functions.
def copy_p2pd_install_files_as_needed():
    # Make install dir if needed.
    install_root = get_p2pd_install_root()
    pathlib.Path(install_root).mkdir(parents=True, exist_ok=True)

    # Copy KVS db if needed.
    script_parent = get_script_parent()
    kvs_db_copy_path = get_kvs_db_copy_path(script_parent)
    kvs_db_install_path = get_kvs_db_install_path(install_root)
    if not os.path.isfile(kvs_db_install_path):
        shutil.copy(kvs_db_copy_path, kvs_db_install_path)

if __name__ == '__main__':
    pass


import asyncio
import io
from struct import unpack
from hashlib import md5
from .utils import *
from .address import *
from .turn_defs import *

# Parse a TURN message.
# Use bitwise OPs to get valid method and status codes.
def turn_parse_msg(buf):
    try:
        turn_msg, _ = TurnMessage.unpack(buf)
        turn_method = b_and(turn_msg.msg_type, b"\x00\x0f")
        turn_status = b_and(turn_msg.msg_type, b"\x01\x10")
        return turn_msg, turn_method, turn_status
    except Exception:
        return None, None, None

"""
Messages sent to a relay address get returned by the TURN
server to the client as a message with a:
A) DATA attribute (the message)
B) Peer Address attribute (the sender)

Return this information to the caller.
"""
def turn_get_data_attr(msg, af):
    # Step through all attributes.
    data = peer_tup = None
    while not msg.eof():
        attr_code, _, attr_data = msg.read_attr()

        # The message segment.
        if attr_code == TurnAttribute.Data:
            if isinstance(attr_data, memoryview):
                data = attr_data.tobytes()
            else:
                data = attr_data

        # The sender of the message.
        if attr_code == TurnAttribute.XorPeerAddress:
            peer_tup = turn_peer_attr_to_tup(
                attr_data,
                msg.txn_id,
                af
            )

    # Reset attribute pointer to start.
    msg.attr_cursor = 0

    # Return results (if any.)
    return data, peer_tup

# True when all the fields in the client needed for auth are set.
def is_auth_ready(self):
    key_con = self.key is not None
    realm_con = self.realm is not None
    nonce_con = self.nonce is not None
    if key_con and realm_con and nonce_con:
        return True
    else:
        return False

# Processes attributes from a TURN message.
async def process_attributes(self, msg):
    # Unpack attributes from message.
    error_code = 0
    error_msg = b''
    while not msg.eof():
        attr_code, _, attr_data = msg.read_attr()

        # Src port of UDP packet + external IP.
        # These details are XORed based on whether its IPv4 or IPv6.
        if attr_code == TurnAttribute.XorMappedAddress:
            if self.mapped != []:
                continue

            self.mapped = turn_peer_attr_to_tup(
                attr_data,
                msg.txn_id,
                self.turn_addr.af
            )
            log("> Turn setting mapped address = {}".format(self.mapped))
            self.client_tup_future.set_result(self.mapped)

        # Server address given back for relaying messages.
        if attr_code == TurnAttribute.XorRelayedAddress:
            if self.relay_tup is not None:
                continue

            # Extract the relay address info to a tup.
            self.relay_tup = turn_peer_attr_to_tup(
                attr_data,
                msg.txn_id,
                self.turn_addr.af
            )

            # Indicate the tup has been set.
            self.relay_tup_future.set_result(self.relay_tup)
            log(f"> Turn setting relay addr = {self.relay_tup}")
            self.relay_event.set()

        # Handle authentication.
        if attr_code == TurnAttribute.Realm:
            self.realm = attr_data
            if self.turn_user is not None and self.turn_pw is not None:
                self.key = md5(self.turn_user + b':' + self.realm + b':' + self.turn_pw).digest()
                log("> Turn setting key = %s" % ( to_s(to_h(self.key)) ) )

        # Nonce is used for reply protection.
        # As our client uses a state-machine the impact of this is minimal.
        elif attr_code == TurnAttribute.Nonce:
            self.nonce = attr_data
            if IS_DEBUG:
                log("> Turn setting nonce = %s" % ( to_s(to_h(self.nonce.tobytes()))  ) )

        elif attr_code == TurnAttribute.Lifetime:
            self.lifetime, = unpack("!I", attr_data)
            if IS_DEBUG:
                log("> Turn setting lifetime = %d" % ( self.lifetime ))

        # Return any error codes.
        elif attr_code == TurnAttribute.ErrorCode:
            b2 = io.BytesIO(attr_data)
            d = b2.read(4)
            error_code = (d[2] & 0x7) * 100 + d[3]
            error_msg = b2.read()

    # Trigger auth ready event.
    if is_auth_ready(self):
        if not self.auth_event.is_set():
            self.auth_event.set()

    # Reset attribute pointer to start.
    msg.attr_cursor = 0

    # Return any errors info.
    return [error_code, error_msg]

# Process any replies from the TURN server.
# This function is run concurrently and doesn't block the main program.
async def process_replies(self):
    # Keep processing until stopped.
    while self.state != TURN_ERROR_STOPPED:
        # Prune old tasks.
        self.tasks = rm_done_tasks(self.tasks)

        # Async wait for up to N seconds for new messages.
        try:
            out = await self.turn_pipe.recv(timeout=1)
        except Exception:
            await asyncio.sleep(1)
            continue

        # Timeout no messages.
        if out is None:
            await asyncio.sleep(1)
            continue

        # Option B) Parse TURN messages from the server.
        turn_msg, turn_method, turn_status = turn_parse_msg(memoryview(out))
        if turn_msg is None:
            continue

        """
        Some TURN messages may have data attributes.
        These indicate a peer who sent data to our relay address.
        Attempt to look for these attributes and process them if found.
        """
        msg_data, peer_tup = turn_get_data_attr(turn_msg, self.turn_addr.af)
        if msg_data is not None and peer_tup is not None:
            # Not a peer we white listed.
            peer_ip = peer_tup[0]
            if peer_ip not in self.peers:
                log("Got a message from an unknown peer.")
                continue

            # Get relay address to route to sender of the message.
            peer_relay_tup = self.peers[peer_ip]

            # Tell the sender that we got the message.
            _, payload = self.stream.handle_ack(
                msg_data,
                self.stream.is_ack,
                self.stream.is_ackable,
                lambda buf: self.stream.send(buf, peer_relay_tup)
            )

            """
            A simple ACK-based protocol is transparently applied to the
            relay messages behind the scenes to add reliability.
            If the header can't be found then the original message
            will be unknown so we skip it.
            """
            if payload is None:
                log(f"Payload from turn was None but msg data = {msg_data}")
                if not self.blank_rudp_headers:
                    continue
                else:
                    self.handle_data(msg_data, peer_relay_tup)
                    continue

            """
            The senders message has been stripped of the ACK header.
            It is then routed to this object (pipe-like object)
            where it will be handled and/or queued. The sender's
            relay address is listed as the sender to make it
            easy to route replies transparently.
            """
            self.handle_data(payload, peer_relay_tup)
            continue
    
        """
        When a TURN message is sent it has a unique TXID.
        Replies in response to these messages use the same TXID.
        Unknown TXIDs for messages are discarded.
        """
        txid = turn_msg.txn_id
        if txid not in self.msgs:
            log("Got turn message with unknown TXID.")
            continue

        # A few important attributes are saved into the client for future use.
        # Mostly details for relaying and authentication.
        try:
            error_code, error_msg = await process_attributes(self, turn_msg)
        except Exception:
            log_exception()
            continue

        # Log any error messages.
        if turn_status == TurnMessageCode.ErrorResp:
            log('Turn error {}: {}'.format(error_code, error_msg))

            # Stale nonce.
            if error_code == 438:
                """
                The client looks for the MESSAGE-INTEGRITY attribute in the response
                (either success or failure).  If present, the client computes the
                message integrity over the response as defined in Section 15.4, using
                the same password it utilized for the request.  If the resulting
                value matches the contents of the MESSAGE-INTEGRITY attribute, the
                response is considered authenticated.  If the value does not match,
                or if MESSAGE-INTEGRITY was absent, the response MUST be discarded,
                as if it was never received.
                """
                log(f"stole nonce. retransmit for {txid}")
                self.msgs[txid]["status"].set_result(STATUS_RETRY)
                continue

        # Attempt to authenticate or create a relay address or refresh one.
        if turn_method == TurnMessageMethod.Allocate:
            log("got alloc")

            # Notify sender that message was received.
            self.msgs[txid]["status"].set_result(STATUS_SUCCESS)
            if turn_status == TurnMessageCode.SuccessResp:
                if self.state != TURN_TRY_ALLOCATE:
                    #self.txid = txid
                    #self.requires_auth = False
                    self.auth_event.set()
                
        

            """
            The first 'allocate' message makes the server return attributes
            needed to authenticate and sign all future messages.
            Hence failing once is expected.
            """
            if turn_status == TurnMessageCode.ErrorResp:
                # Avoid infinite loop of allocations.
                if self.state != TURN_TRY_ALLOCATE:
                    self.set_state(TURN_TRY_ALLOCATE)

                    # All future messages from here-on in are 'signed.'
                    task = asyncio.create_task(
                        async_retry(
                            lambda: self.allocate_relay(sign=True),
                            count=5
                        )
                    )
                    self.tasks.append(task)

            continue

        # White list a particular peer to send replies to our relay address.
        if turn_method == TurnMessageMethod.CreatePermission:
            if turn_status == TurnMessageCode.SuccessResp:
                # Notify sender that message was received.
                self.msgs[txid]["status"].set_result(STATUS_SUCCESS)

            continue

        if turn_method == TurnMessageMethod.Refresh:
            self.msgs[txid]["status"].set_result(STATUS_SUCCESS)
            continue

    self.turn_client_stopped.set()


from p2pd.utils import async_test
from p2pd.cmd_tools import *
from p2pd.net import *
from p2pd.netiface_extra import *
from p2pd.interface import *

async def do_stuff():
    #netifaces = await Netifaces().start()
    print(netifaces.interfaces())
    if_name = netifaces.interfaces()[0]
    print(netifaces.ifaddresses(if_name))
    
    return
    i = await Interface(if_name).start()
    print(i)

async_test(do_stuff)


"""
nc -4 -u p2pd.net 7

"""

from p2pd.test_init import *
from p2pd.utils import log_exception, what_exception
from p2pd import STUNClient, Interface
from p2pd.net import VALID_AFS, TCP, UDP
from p2pd.base_stream import pipe_open, SUB_ALL, BaseProto
from p2pd.stun_client import tran_info_patterns, do_stun_request
from p2pd.stun_client import changeRequest, changePortRequest
from p2pd.ip_range import IPRange

class TestIPv6(unittest.IsolatedAsyncioTestCase):
    async def test_ipv6(self):
        """
        i = await Interface().start_local()
        s = STUNClient(interface=i, proto=UDP, af=IP6)
        x = await s.get_wan_ip()
        print(x)
        """

        netifaces = await init_p2pd()
        i_name = netifaces.interfaces()[0]
        addr_info = netifaces.ifaddresses(i_name)
 
        servers = [
            {
                "host": "p2pd.net",
                "primary": {"ip": "2a01:4f9:3081:50d9::2", "port": 3478},
                "secondary": {"ip": "2a01:4f9:3081:50d9::3", "port": 3479},
            },

        ]

        i = Interface()
        await i.start()

        print(i)
        return



        s = STUNClient(interface=i, proto=UDP, af=IP6)
        x = await s.get_wan_ip(servers=servers)
        print("wan ip == ")
        print(x)
        return




        ipr_nic = IPRange("fe80::f75c:ff8e:b0dd:2ce2")
        ipr_ext = IPRange("2403:5800:b018:8000::4a0d")
        route = Route(
            IP6,
            nic_ips=[ipr_nic],
            ext_ips=[ipr_ext],
            interface=i
        )

        await route

        dest = await Address("p2pd.net", 7, route)
        pipe = await pipe_open(TCP, route, dest)
        await pipe.send(b"Test")
        out = await pipe.recv()
        print(out)
        await pipe.close()
                

if __name__ == '__main__':
    main()


"""
This module is a drop-in replacement for netifaces on Windows.
Usage is simply:
    from p2pd import *
    async def main():
        netifaces = await init_p2pd() 

The pypi netifaces module has several problems on Windows OS':

1. It requires a distribution of VS C++ Build Tools 20**.
Having users install the right build tools to get the software
working is complex and error-prone. It also makes packaging
software that uses netifaces much more difficult.

2. On recent versions of Windows it displays meaningless
strings of GUIDs over network interface names. Consequently,
one has to do a series of reg hacks, privilege elevations,
and patches to use the library with human-readable names.

The most portable way to do complex operations on Windows
seems to be to use powershell scripting. Looking up certain
registry keys isn't guaranteed to work as the locations may
change between versions. The networking tools available in
cmd.exe are now being phased out in favor of the tools
available in powershell. Powershell is now widely supported
even on older Windows OS.

The downside to using powershell to obtain relevant NIC
information is it's slow. A process needs to be spawned
for each new command. In order to prevent errors on certain
Windows versions concurrency also has to be disabled. This
doesn't seem to effect speed too much but the code is many
times slower than netiface. Nevertheless, it's more
portable and doesn't require privilege elevations to get
human-readable network interface descriptions.

Speedups:

I've added a new feature to obtain all the networking info
from a single powershell script. The program will attempt
to use this script if powershell is unrestricted. If it
fails to load interfaces with regular commands it will
try relaunch with a UAC prompt, unrestrict powershell,
then attempt to run the script in powershell.

Registry:

Don't attempt to write a new version of this using the
Windows registry. The registry does not make available
simple access to the interface numbers and a special
(undocumented) algorithm needs to be used to rank the
interfaces and derive the number. Doing this is unsupported
outside of using heavy Windows SDK dependencies.

Notes:
    - These commands don't seem to require special permissions.
    - They need to use double quotes or the command won't run.
    - Tested as working with execution policy = restricted.
    - Is there a way to convert the PS1 script to a single line
    command and have it passed to powershell without execution perms?
"""

import re
import base64
from .ip_range import *
from .cmd_tools import *

CMD_TIMEOUT = 10

IFS_PS1 = """
# Load default interface for IPv4 and IPv6.
$v4_default = Find-NetRoute -RemoteIPAddress 0.0.0.0 -erroraction 'silentlycontinue' | Format-List -Property ifIndex
$v6_default = Find-NetRoute -RemoteIPAddress :: -erroraction 'silentlycontinue' | Format-List -Property ifIndex
if($v4_default -eq $null){
    $v4_default = "null"
}
if($v6_default -eq $null){
    $v6_default = "null"
}

# Show them if any.
echo("4444444444")
echo($v4_default)
echo("4444444444")
echo("6666666666")
echo($v6_default)
echo("6666666666")

# Load interfaces and associated addresses. removed -physical
$ifs = Get-NetAdapter -erroraction 'silentlycontinue' | where status -eq up 
Foreach($iface in $ifs){
    # Get first hop for the iface for both AFs.
    $v4gw = (Get-NetRoute "0.0.0.0/0" -InterfaceIndex $iface.ifIndex  -erroraction 'silentlycontinue').NextHop 
    $v6gw = (Get-NetRoute "::/0" -InterfaceIndex $iface.ifIndex  -erroraction 'silentlycontinue').NextHop
    if($v4gw -eq $null){
        $v4gw = "null"
    }

    if($v6gw -eq $null){
        $v6gw = "null"
    }

    # Build a list of IPs.
    $ips = [System.Collections.ArrayList]::new()
    $addrs = Get-NetIPAddress -InterfaceIndex $iface.ifIndex
    Foreach($addr in $addrs){
        $ips += $addr.IPAddress
    }

    # Save them as new property values.
    $iface | Add-Member -NotePropertyName v4GW -NotePropertyValue $v4gw
    $iface | Add-Member -NotePropertyName v6GW -NotePropertyValue $v6gw

    # Output this interface info with it's addresses.
    $out = $iface | Format-List -Property InterfaceDescription,ifIndex,InterfaceGuid,MacAddress,v4GW,v6GW
    echo($out)
    echo($ips)
}
"""

async def load_ifs_from_ps1():
    # Get all interface details as one big script.
    out = await ps1_exec_trick(IFS_PS1)

    # Load default interface by if_index.
    default_ifs = {IP4: None, IP6: None}
    if_defaults_by_index = {}
    for v in [4, 6]:
        # No to AF.
        af = v_to_af(v)

        # Regex to extract the interface no for the AF.
        delim = str(v) * 10
        p = f"{delim}[\s\S]*ifIndex\s*:\s*([0-9]+)[\s\S]*{delim}"
        if_index = re.findall(p, out)

        # Save it in a loopup table.
        if len(if_index):
            if_index = to_n(if_index[0])

            # Save it by AF.
            default_ifs[af] = if_index

            # Make a list to store AF by if_index.
            if if_index not in if_defaults_by_index:
                if_defaults_by_index[if_index] = []

            # Save AF by if_index.
            if_defaults_by_index[if_index].append(af)

    # Extract interface details.
    p = "InterfaceDescription *: *([^\r\n]+?) *[\r\n]+ifIndex *: *([0-9]+?) *[\r\n]+InterfaceGuid *: *([^\r\n]+?) *[\r\n]+MacAddress *: *([^\r\n]+?) *[\r\n]+v4GW *: *([^\r\n]+?) *[\r\n]+v6GW *: *([^ ]+?)[\s]+((?:[0-9a-f.:%]+ *[\r\n]*)+)"
    re_results = re.findall(p, out)

    # Index the results into a dict.
    if_infos = []
    if len(re_results):
        for r in re_results:
            if_index = to_n(r[1])
            if_info = {
                "guid": r[2],
                "name": r[0],
                "no": if_index,
                "mac": r[3],

                # Placeholders.
                "addr": None,
                "gws": { 
                    IP4: None if r[4] == "null" else r[4],
                    IP6: None if r[5] == "null" else r[5]
                },
                "defaults": []
            }

            # Set defaults.
            if if_index in if_defaults_by_index:
                if_info["defaults"] = if_defaults_by_index[if_index]

            # Process address info.
            addr_info = { IP4: [], IP6: [] }
            addr_s = r[6].replace(' ', '')
            addr_list = addr_s.splitlines(False)
            for addr in addr_list:
                # Skip blank IPs.
                if addr == '':
                    continue

                # Convert to netifaces format.
                addr = ip_strip_cidr(ip_strip_if(addr))
                ipr = IPRange(addr)
                addr_info[ipr.af].append({
                    "addr": addr,
                    "af": ipr.af,
                    "cidr": ipr.cidr,
                    "netmask": ipr.netmask
                })

            # Save addresses.
            if_info["addr"] = addr_info

            # Save info.
            if_infos.append(if_info)

    """
    If no route found for a given address family set the first
    interface as the default interface for that AF.
    """
    for af in VALID_AFS:
        if default_ifs[af] is None:
            if len(if_infos):
                if if_infos[0]["gws"][af] is not None:
                    if_infos[0]["defaults"].append(af)


    return if_infos

async def get_default_gw_by_if_index(af, if_index):
    dest_ip = "0.0.0.0/0" if af == IP4 else "::/0"
    cmd_str = '{} "(Get-NetRoute {} -InterfaceIndex {}).NextHop"'
    cmd_str = cmd_str.format("powershell", dest_ip, if_index)

    # Execute the command.
    try:
        out = await cmd(cmd_str, timeout=CMD_TIMEOUT)
    except Exception:
        return None


    if out is None:
        return None

    # Return the string if it's a valid IP.
    out = out.strip()
    try:
        ip_f(out)
        return out
    except:
        log(f"{out}")
        return None

async def get_addr_info_by_if_index(if_index):
    addr = { IP4: [], IP6: [] }
    cmd_str = 'powershell "Get-NetIPAddress -InterfaceIndex {}"'
    cmd_str = cmd_str.format(if_index)
    try:
        out = await cmd(cmd_str, timeout=CMD_TIMEOUT)
    except Exception:
        return addr

    try:
        addr_infos = re.findall(
            "IPAddress\s*:\s*([^\s]*)[\s\S]*?AddressFamily\s*:\s*([^\s]+)[\s\S]*?PrefixLength\s*:\s([0-9]+)",
            out
        )

        for addr_info in addr_infos:
            ip_val, af_family, cidr = addr_info
            cidr = int(cidr)
            if af_family == "IPv4":
                af = IP4
            if af_family == "IPv6":
                af = IP6

            addr[af].append({
                "addr": ip_val,
                "af": af,
                "cidr": cidr
            })
    except Exception:
        log_exception()
        return addr

    return addr

async def get_default_iface_by_af(af):
    if af == IP4:
        any_offset = 0
    if af == IP6:
        any_offset = 1

    any_addr_list = ["0.0.0.0", "::"]
    dest_ip = any_addr_list[any_offset]
    cmd_buf = 'powershell "Find-NetRoute -RemoteIPAddress {}"'
    cmd_buf = cmd_buf.format(dest_ip)
    try:
        out = await cmd(cmd_buf, timeout=CMD_TIMEOUT)
    except:
        return None

    try:
        if_index_str = re.findall("InterfaceIndex\s*:\s*([0-9]+)", out)
        if len(if_index_str):
            return int(if_index_str[0])
        else:
            # If an AF is not support an error is thrown
            # and the pattern above won't match anything.
            return None
    except Exception:
        log_exception()
        return None

def extract_if_fields(ifs_str):
    results = []
    try:
        if_info_matches = re.findall("InterfaceDescription\s*:\s([^\r\n]*?)[\r\n]+ifIndex\s*:\s*([0-9]+)\s*InterfaceGuid\s*:\s*([^\r\n]+)\s*MacAddress\s*:\s*([^\s]+)\s*", ifs_str)
        if len(if_info_matches):
            for if_info_match in if_info_matches:
                if_desc, if_index, guid, mac_addr = if_info_match
                if_index = int(if_index)
                results.append({
                    "guid": guid,
                    "name": if_desc,
                    "no": if_index,
                    "mac": mac_addr,

                    # Placeholders.
                    "addr": None,
                    "gws": { IP4: None, IP6: None },
                    "defaults": None
                })
    except Exception:
        log_exception()
        return results

    return results

# Get list of net adaptors via powershell.
# Ignore hidden adapters. Non-physical or down.
# Specify desc and index to show full entry.
async def get_ifaces():
    try:
        out = await cmd('powershell "Get-NetAdapter -physical | where status -eq up  | Format-List -Property InterfaceDescription,ifIndex,InterfaceGuid,MacAddress"', timeout=CMD_TIMEOUT)
    except Exception:
        log_exception()
        out = ""

    return out

async def win_load_interface_state(if_results):
    # Lookup whether an1
    if_defaults_by_index = {}
    af_index = {}
    async def set_ip4_if_index():
        af_index[IP4] = await get_default_iface_by_af(IP4)
    async def set_ip6_if_index():
        af_index[IP6] = await get_default_iface_by_af(IP6)

    # Execute the above functions.
    tasks =  [
        set_ip4_if_index(),
        set_ip6_if_index(),
    ]
    for task in tasks:
        await task

    # Set the AFs the interface is the default for.
    if af_index[IP4] or af_index[IP6]:
        if af_index[IP4]:
            if_defaults_by_index[af_index[IP4]] = [IP4]

        if af_index[IP6]:
            if_defaults_by_index[af_index[IP6]] = [IP6]

        if af_index[IP4] == af_index[IP6]:
            if_defaults_by_index[af_index[IP4]] = [IP4, IP6]

    # Parse output lines from Get-Adapter.
    if_tasks = []
    by_guid_index = {}
    for result in if_results:
        async def if_task_func():
            try:
                if_index = result["no"]
                guid = result["guid"]

                # Default interface for these address families.
                default_for_afs = []
                if if_index in if_defaults_by_index:
                    default_for_afs = if_defaults_by_index[if_index]

                result["defaults"] = default_for_afs

                # Otherwise it's worth saving.
                by_guid_index[guid] = result

                # Get address information.
                async def set_addr():
                    by_guid_index[guid]["addr"] = await get_addr_info_by_if_index(if_index)

                # Get default gateways.
                async def set_ip4_gw():
                    by_guid_index[guid]["gws"][IP4] = await get_default_gw_by_if_index(IP4, if_index)
                async def set_ip6_gw():
                    by_guid_index[guid]["gws"][IP6] = await get_default_gw_by_if_index(IP6, if_index)

                # Execute the above tasks.
                sub_tasks = [
                    set_addr(),
                    set_ip4_gw(),
                    set_ip6_gw()
                ]
                for task in sub_tasks:
                    await task
            except Exception:
                log_exception()
                return

        if_tasks.append(if_task_func())

    for if_task in if_tasks:
        await if_task

    return by_guid_index

def win_set_gateways(by_guid_index):
    gws = {
        "default": { },
        int(IP4): [],
        int(IP6): []
    }

    for _, addr_info in by_guid_index.items():
        # Set defaults for different AFs.
        for af in addr_info["defaults"]:
            # If no gateway for af found skip.
            if addr_info["gws"][af] is None:
                continue

            # Default field for gateway.
            gws["default"][int(af)] = (
                addr_info["gws"][af],
                addr_info["name"]
            )

            # Add to list of gateways.
            if len(addr_info["gws"][af]):
                is_default = af in addr_info["defaults"]
                gws[int(af)].append(
                    (
                        addr_info["gws"][af],
                        addr_info["name"],
                        is_default
                    )
                )

    return gws

class Netifaces():
    AF_INET = IP4
    AF_INET6 = IP6
    AF_LINK = 18

    def __init__(self):
        pass

    async def start(self):
        # Use a single script to load all info at once.
        if_infos = await load_ifs_from_ps1()
        self.by_guid_index = {}
        for if_info in if_infos:
            self.by_guid_index[if_info["guid"]] = if_info

        # Sanity check.
        if self.by_guid_index == {}:
            raise Exception("Unable to load interfaces.")

        # Setup name index.
        self.by_name_index = {}
        for _, if_info in self.by_guid_index.items():
            name = if_info["name"]
            self.by_name_index[name] = if_info

        # Save main gateways used.
        self.gws = win_set_gateways(self.by_guid_index)
        return self

    def gateways(self):
        return self.gws

    def if_info(self, if_name):
        return self.by_name_index[if_name]

    def guid(self, if_name):
        if_info = self.if_info(if_name)
        return if_info["guid"]

    def nic_no(self, if_name):
        if_info = self.if_info(if_name)
        return if_info["no"]

    def ifaddresses(self, if_name):
        if_info = self.by_name_index[if_name]
        addr_format = {
            int(IP4): [],
            int(IP6): [],

            # Netifaces AF_LINK = MAC address.
            Netifaces.AF_LINK: [
                {
                    "addr": if_info["mac"]
                }
            ]
        }

        # Add addresses in netiface format.
        for af in [IP4, IP6]:
            for addr_info in if_info["addr"][af]:
                addr = {
                    "addr": addr_info["addr"],
                    "netmask": cidr_to_netmask(
                        addr_info["cidr"],
                        af
                    )
                }

                addr_format[int(af)].append(addr)

        return addr_format

    def interfaces(self):
        ifs = []
        for _, if_info in self.by_guid_index.items():
            ifs.append(if_info["name"])

        return ifs


import hashlib
import secrets
from p2pd import *

COMPUTER_A_NAME = ["computer_a", ".node"]

# Put your custom protocol code here.
async def msg_cb(msg, client_tup, pipe):
    # E.G. add a ping feature to your protocol.
    if b"PING" in msg:
        await pipe.send(b"PONG")

# Computer a and b code can run on different
# computers -- for demo they're just on the same.
async def computer_a():
    # Start our main node server.
    # The node implements your protocol.
    node = await start_p2p_node(
        # Used to create the accounts that can modify COMPUTER_A_NAME!
        # Save your seed value to reuse it! Otherwise names are lost.
        seed=hashlib.sha3_256(b"computer a unique password"),
        #
        # Set to true for port forwarding + pin holes.
        enable_upnp=False,
        #
        # Make sure node server uses different port.
        port=NODE_PORT + 50 + 1
        
    )
    node.add_msg_cb(msg_cb)
    #
    # Register a human readable name for this peer.
    # NOTE: for demo only -- use your own unique name!
    await node.register(COMPUTER_A_NAME)
    #
    return node

async def computer_b():
    # Start our main node server.
    # The node implements your protocol.
    node = await start_p2p_node(
        # Used to create the accounts that can modify computer b's names!
        # Save your seed value to reuse it! Otherwise names are lost.
        seed=secrets.token_bytes(24),
        #
        # Set to true for port forwarding + pin holes.
        enable_upnp=False,
        #
        # Make sure node server uses different port
        # to computer_a.
        port=NODE_PORT + 50 + 2
    )
    #
    # Spawn a new pipe from a P2P con.
    # Connect to their node server.
    pipe, success_type = await node.connect(COMPUTER_A_NAME)
    #
    # Test send / receive.
    msg = b"test send"
    await pipe.send(b"ECHO " + msg)
    out = await pipe.recv()
    #
    # Cleanup.
    assert(msg in out)
    await pipe.close()
    #
    return node
    
# Warning: startup is slow - be patient.
async def example():
    """
    (1) Computer A starts a node server and uses 'IRCDNS'
    to store its address at a given name.
    """
    node_a = await computer_a()
    #
    #
    """
    (2) Computer B starts a node server and uses 'IRCDNS'
    to lookup the p2p address of computer a to connect to it.
    """
    node_b = await computer_b()
    #
    #
    # Cleanup / shut down node servers.
    await node_a.close()
    await node_b.close()

# Run the coroutine.
# Or await example() if in async REPL.
if __name__ == '__main__':
    async_test(example)


import asyncio
import struct
import random
from struct import pack
from .utils import *

UDP_MAX_DICT_LEN = 1000

"""
Extended functionality to allow the UDP stream class
to provide 'reliable' packet delivery. It uses message
IDs for each message and acknowledgements. It doesn't
guarantee ordered delivery. Inherited by udp_stream.
"""

class ACKUDP():
    def __init__(self):
        self.seq = {} # Waiting for acks.
        self.ack_send_tasks = []

    # Returns a sequence number if a message is an ack.
    def is_ack(self, data, stream):
        if len(data) >= 9:
            seq, = struct.unpack("!Q", data[0:8])
            is_ack = data[8]
            if is_ack == 1:
                return seq

        return None

    # Received message that needs to be acked.
    # Return its sequence number and valid ack response.
    def is_ackable(self, data, stream):
        payload = ack = is_ack = seq = None
        if len(data) >= 9:
            seq, = struct.unpack("!Q", data[0:8])
            is_ack = data[8]
        else:
            return [None, None, None]

        if is_ack == 0:
            # Build ack message to send in response.
            ack = struct.pack("!Q", seq) + struct.pack("!B", 1)

        return [seq, ack, data[9:]]

    """
    Clients that receive a message that can be 'acked' now
    send back the ack every time they receive a message even
    if they have already acked. This makes more sense as we
    don't know if the receiver has actually gotten the ack
    yet. Keep code to skip acking if a peer sent a message.
    This prevents getting into loops for the sender.
    """
    def handle_ack(self, data, f_is_ack, f_is_ackable, f_send):
        self.ack_send_tasks = rm_done_tasks(self.ack_send_tasks)
        data = data
        payload = recv_seq = ack_seq = ack = None
        self.timestamp = timestamp()

        # If this message is an ack then record its seq no.
        if f_is_ack is not None:
            ack_seq = f_is_ack(data, self)
            if ack_seq is not None:
                if ack_seq in self.seq:
                    self.seq[ack_seq].set()

                return 0, payload

        # If it's a regular message check if it needs
        # to be acknowledged and record the seq no.
        if f_is_ackable is not None:
            recv_seq, ack, payload = f_is_ackable(data, self)
            if payload is None:
                return 0, None

            if ack is not None:
                # Seq is set when sending.
                # If they give us back our seq take it as an ACK
                # even if they didn't set the ACK flag.
                if recv_seq in self.seq:
                    # Pretend we received an ACK for our message.
                    self.seq[recv_seq].set()

                    # Don't broadcast an ACK for this.
                    ack = None

        # Keep dicts from taking up too much memory.
        if len(self.seq) > UDP_MAX_DICT_LEN:
            self.seq = {}

        """
        The TURN client implements a custom is_ackable that wraps an ACK
        in a channel message which allows the server to deliver the message.
        """
        if ack is not None:
            task = asyncio.create_task(
                async_wrap_errors(
                    f_send(ack)
                )
            )

            self.ack_send_tasks.append(task)
            return 2, payload

        return 1, payload

    """
    A function that retransmits a UDP packet up to 'tries' time or
    'sock_timeout' duration. If a special acknowledgement is received
    before an error condition - the function returns successfully with
    a value of 0 (no errors.) The code uses events to wait on ACKs
    so there are no inefficient busy-loop checks.
    """
    async def ack_send(self, data, dest_tup, seq=None, sock_timeout=0, tries=3):
        # Keep sending until max sends reached.
        # For acks we send max transmits as they're small messages.
        if seq is None:
            seq = random.randrange(1, (2 ** (8 * 8)))

        """
        Mark all messages we send in the same data structure clients
        use to indicate whether they have acknowledged a message.
        This prevents the sender from getting into loops.
        """
        event = asyncio.Event()
        self.seq[seq] = event

        # Do the sending concurrently so event can be returned.
        async def worker():
            # Record when the process started.
            start = 0
            if sock_timeout:
                start = timestamp()

            # Build data to send.
            buf = bytearray().join([
                pack("!Q", seq),
                pack("!B", 0),
                memoryview(data)
            ])

            # Await on ACK events.
            # Break on transmits >= tries, timeout, or success.
            send_transmits = 0
            while True:
                # Initial send.
                await self.send(buf, dest_tup)
                send_transmits += 1

                # Finish trying to send.
                # First failure mode reached.
                if send_transmits >= tries:
                    break

                # Recheck for ack every second.
                try:
                    # Will return instantly on receiving a related ACK.
                    # Otherwise it suspends for other code to execute.
                    await asyncio.wait_for(
                        self.seq[seq].wait(),
                        3
                    )

                    # No timeout error = success.
                    break
                except asyncio.TimeoutError:
                    pass

                # Too much time passed.
                # Second failure mode reached.
                if sock_timeout:
                    elapsed = timestamp() - start
                    if elapsed >= sock_timeout:
                        break

            # Do cleanup.
            if seq in self.seq:
                del self.seq[seq]

        # Schedule sending task.
        task = asyncio.create_task(worker())
        self.ack_send_tasks.append(task)

        # Wait for ACK.
        return task, event

class BaseACKProto(asyncio.Protocol):
    def __init__(self, conf):
        self.conf = conf

    # Supports dropping duplicate messages.
    def is_unique_msg(self, pipe, data, client_tup):
        # Reset seen msgs after dict fills.
        if len(self.msg_ids) > self.conf["max_msg_ids"]:
            self.msg_ids = {}

        # Record msg -- drop if already seen.
        # Route by client endpoint.
        # Seen messages are per client IP.
        buf = to_b(client_tup[0]) + data

        """
        I use Pythons insecure hash function.
        A cryptographically secure hash func
        would absolutely destroy the event
        loops performance! E.g. 100 ms+ per hash,
        per message received = yikes.
        """
        msg_id = hash(buf)
        if msg_id in self.msg_ids:
            return 0
        else:
            self.msg_ids[msg_id] = 1
            return 1


from p2pd.test_init import *
import random
from p2pd.clock_skew import *
from p2pd.settings import *

class TestClockSkew(unittest.IsolatedAsyncioTestCase):
    async def test_get_ntp_pool_ntp(self):
        i = await Interface().start_local()
        ntp = None

        for _ in range(0, 5):
            server = random.choice(NTP_SERVERS)
            ntp = await get_ntp(i, server=server[0])
            if ntp:
                break

        self.assertTrue(ntp)

    async def test_get_clock_skew(self):
        i = await Interface().start_local()
        sys_clock = await SysClock(i).start()
        self.assertTrue(sys_clock.clock_skew)

if __name__ == '__main__':
    main()


from p2pd.test_init import *
from p2pd import *
import re

"""
All of the Python code used in the documentation gets tested
using this module. Easy to know if examples still work.
"""

EXAMPLES_DIR = "../docs/source/python/examples"

class TestPyExamples(unittest.IsolatedAsyncioTestCase):
    async def do_py_example(self, n):
        print(n)
        with open(f"{EXAMPLES_DIR}/example_{n}.py") as fp:
            py_code = fp.read()

            # Event loop is already running so replace
            # async_test with an await call.
            py_code = py_code.replace(
                'async_test(example)',
                'pass'
            )

            # Load the example code definitions.
            try:
                exec(py_code, globals())
            except Exception as e:
                print(f"Py example {n} failed.")
                what_exception()
                assert(0)

            # Run the async callback.
            coro = globals().get("example")
            await coro()

        log(f"py example {n} passed")

    # So they can be tested concurrently.
    async def test_1(self):
        print("Skipping testing py ex 1 to avoid registering IRC accounts.")
        #await self.do_py_example(1)

    async def test_2(self):
        await self.do_py_example(2)

    async def test_3(self):
        await self.do_py_example(3)

    async def test_4(self):
        await self.do_py_example(4)

    async def test_5(self):
        await self.do_py_example(5)

    async def test_6(self):
        await self.do_py_example(6)

    async def test_7(self):
        await self.do_py_example(7)

    async def test_8(self):
        await self.do_py_example(8)

    async def test_9(self):
        await self.do_py_example(9)

    async def test_10(self):
        await self.do_py_example(10)

    async def test_11(self):
        await self.do_py_example(11)

    async def test_12(self):
        await self.do_py_example(12)

    async def test_13(self):
        await self.do_py_example(13)

    async def test_14(self):
        await self.do_py_example(14)

    async def test_15(self):
        await self.do_py_example(15)

    async def test_16(self):
        await self.do_py_example(16)

    async def test_17(self):
        await self.do_py_example(17)
    

if __name__ == '__main__':
    main()


from p2pd.test_init import *
from p2pd import IPRange, HOST_TYPE_IP, HOST_TYPE_DOMAIN
from p2pd import IP6, IP4, Route, Interface
from p2pd.net import VALID_AFS, ip_norm, TCP, UDP, Bind
from p2pd.echo_server import EchoServer
from p2pd.base_stream import pipe_open, SUB_ALL
from p2pd.address import Address
from p2pd.utils import what_exception
from p2pd.daemon import Daemon

asyncio.set_event_loop_policy(SelectorEventPolicy())
class TestDaemon(unittest.IsolatedAsyncioTestCase):
    async def test_listen_all_interface_target(self):
        i = await Interface().start_local()
        i.rp[IP6].routes = []

        # Daemon instance.
        server_port = 10126
        proto = TCP
        echod = await EchoServer().listen_all(
            [i],
            [server_port],
            [proto]
        )

        await echod.close()

    async def test_bind_target(self):
        d = Daemon()
        i = await Interface().start_local()
        i.rp[IP6].routes = []

        p = 10233
        b = await Bind(i, i.supported()[0], port=p).bind(p)
        await d._listen(
            target=b,
            port=p,
            proto=TCP
        )

        await d.close()

    async def test_listen_specific(self):
        d = Daemon()
        i = await Interface().start_local()
        i.rp[IP6].routes = []

        p = 10233
        b = await Bind(i, i.supported()[0], port=p).bind(p)
        await d.listen_specific(
            targets=[[b, TCP]],
        )

        await d.close()

    async def test_daemon(self):
        server_port = 0
        loopbacks = {
            IP4: "127.0.0.1",
            IP6: "::1"
        }

        at_least_one = False
        for af in [IP4, IP6]:
            try:
                interface = await Interface(af).start()
                at_least_one = True
            except Exception:
                """
                Skip unsupported AF types. Meaning they cannot use
                the Internet with that address family. Will ensure
                IP6 has at least one global IP and one link local.
                """
                continue

            """
            (1) Get first route for AF type.
            (2) Use in-built method and manually specify bind IP of '*'.
            (3) For IP4 this will bind to 0.0.0.0.
            (4) For IP6 this will bind to ::.
            (5) Will test IP4 NIC IPs, IP6 link locals / local host.
            """
            route = await interface.route(af)
            addrs = [route.nic(), loopbacks[af]]


            # Test connect to link local.
            if af == IP6:
                addrs.append(route.link_local())

            for addr in addrs:
                msg = b"hello world ay lmaoo"
                for proto in [UDP, TCP]:
                    # Fresh route per server.
                    route = await interface.route(af).bind(ips=addr, port=server_port)

                    # Daemon instance.
                    echod = await EchoServer().listen_all(
                        [route],
                        [server_port],
                        [proto]
                    )
                    server_port = echod.get_listen_port()
                    dest = await Address(addr, server_port, route).res()

                    # Spawn a pipe to the echo server.
                    test_route = await interface.route(af).bind(ips=addr)
                    pipe = await pipe_open(
                        proto,
                        test_route,
                        dest
                    )
                    self.assertTrue(pipe is not None)

                    # Indicate to save all messages to a queue.
                    pipe.subscribe(SUB_ALL)

                    # Send message to server.
                    #print(dest.tup in pipe.stream.handle)
                    send_ret = await pipe.send(msg, dest.tup)

                    # Receive data back.
                    data = await pipe.recv(SUB_ALL)
                    self.assertEqual(data, msg)

                    # Test accept() await.
                    # Send message from pipe to server's client pipe.
                    # Then manually call it's receive and check for receipt.
                    client_pipe = await pipe
                    self.assertTrue(client_pipe is not None)
                    client_pipe.subscribe(SUB_ALL)
                    await pipe.send(msg, dest.tup)
                    data = await client_pipe.recv(SUB_ALL)
                    self.assertEqual(data, msg)

                    """
                    Making sure cleanup works correctly is very important
                    because if they restart a server program it will
                    most probably try listen to the same address and that
                    will throw an 'address already in use' error if the
                    socket wasn't cleaned up correctly. The code here
                    will fail if cleanup for these servers isn't correct.
                    """
                    if pipe is not None:
                        await pipe.close()
                    if echod is not None:
                        await echod.close()

        self.assertTrue(at_least_one)
        await asyncio.sleep(0.1)

if __name__ == '__main__':
    main()

"""
echo server tcp stream socket not being closed.
"""


# Defines all custon exceptions.

# There's no gateway defined for that address family.
class NoGatewayForAF(Exception):
    pass

class InterfaceNotFound(Exception):
    pass

class InterfaceInvalidAF(Exception):
    pass